# -*- coding: utf-8 -*-
"""
Created on Fri Jul  7 16:34:30 2023

@author: dvezinet
"""

import itertools as itt


import numpy as np
from scipy.interpolate import interp1d
import datastock as ds
import matplotlib.pyplot as plt
import matplotlib.lines as mlines


from . import _rockingcurve as rc


__all__ = [
    'scope_reflected_intensities',
]


_DMARKER = {
    'Germanium': {'marker': 'o', 'ms': 8, 'mfc': 'None', 'lw': 2},
    'Quartz': {'marker': 'x', 'ms': 8, 'lw': 2},
}


###########################################################
#
#           Scope over integrated reflectivity
#
###########################################################


def scope_reflected_intensities(
    miller_max=None,
    target_lamb=None,
    materials=None,  
    # plotting
    dmarker=None,
    tit=None,
):
    
    """
    
    Provide target_lamb in m
    
    """

    # -------------------
    # check input
    # -------------------
    
    # miller_max
    miller_max = int(ds._generic_check._check_var(
        miller_max, 'miller_max',
        types=(int, float),
        default=10,
        sign='>0',
    ))
    
    # materials
    lok = ['Quartz', 'Germanium']
    if isinstance(materials, str):
        materials = [materials]
    materials = list(ds._generic_check._check_var_iter(
        materials, 'materials',
        types=(list, tuple),
        types_iter=str,
        default=lok,
        allowed=lok,
    ))
    
    # target_bragg
    target_lamb = ds._generic_check._check_flat1darray(
        target_lamb, 'target_lamb',
        sign='>0',
    )
    
    # dmarker
    if dmarker is None:
        dmarker = _DMARKER
        
    # tit
    tit = ds._generic_check._check_var(
        tit, 'tit',
        types=str,
        default="Scoping integrated reflectivity",
    )
    
    # -------------------
    # Prepare
    # -------------------
    
    ddd = miller_max + 1
    
    dout = {
        'materials': materials,
        'target_lamb': target_lamb,
        'miller_max': miller_max,
        'reflectivity': {
            k0: {
                k1: {
                    'Bragg': np.full((ddd, ddd, ddd), np.nan),
                    'Rint':  np.full((ddd, ddd, ddd), np.nan),
                    'miller': np.full((ddd, ddd, ddd), 'hkl', dtype='S3')
                }
                for k1 in target_lamb
            }
            for k0 in materials
        },
    }    

    # miller indices ranges
    rangeh = np.arange(0, miller_max + 1)
    rangek = np.arange(0, miller_max + 1)
    rangel = np.arange(0, miller_max + 1)

    # -------------------
    # computing
    # -------------------
    
    # Loop over materials
    for lamb0, mat in itt.product(target_lamb, materials):
        results = dout['reflectivity'][mat][lamb0]
    
        # Loop over Miller indices
        for hh, kk, ll in itt.product(rangeh, rangek, rangel):
        
            # skip bulk reflection
            if hh == 0 and kk == 0 and ll == 0:
                continue
 
            # Builds material matrix
            dcry4 = {
                'material': mat,
                'name': 'SCOPING',
                'symbol': f"{mat[:2]}{hh}{kk}{ll}",
                'miller': np.r_[hh, kk, ll],
                'target': {
                    # 'ion': 'Kr34+',
                    'lamb': lamb0 * 1e10,
                    'units': 'A',
                    },
                'd_hkl': None,
            }
 
            results['miller'][hh, kk, ll] = f"{hh}{kk}{ll}"
 
            # Calculates rocking curve
            try:
                
                dout4 = rc.compute_rockingcurve(
                    crystal=dcry4['name'],
                    din=dcry4,
                    lamb=dcry4['target']['lamb'], 
                    plot_power_ratio=False,
                )
    
                results['Bragg'][hh,kk,ll] = dout4['Bragg angle of reference (rad)']*180/np.pi
    
                results['Rint'][hh,kk,ll] = np.trapz(
                    np.nanmean(
                        [
                            dout4['Power ratio'][0,0,0,:],
                            interp1d(
                                dout4['Glancing angles'][1,0,0,:],
                                dout4['Power ratio'][1,0,0,:],
                                bounds_error=False,
                                fill_value = (
                                    dout4['Power ratio'][1,0,0,0],
                                    dout4['Power ratio'][1,0,0,-1],
                                )
                            )(dout4['Glancing angles'][0,0,0,:]),
                        ],
                        axis=0,
                    ),
                    dout4['Glancing angles'][0,0,0,:],
                )

            except:
                pass
 
    # -------------------
    # plotting
    # -------------------
        
    # figure
    fig = plt.figure(figsize=(12, 8))
    fig.suptitle(tit, size=14, fontweight='bold')
    
    # axes
    ax = fig.add_axes([0.1, 0.1, 0.7, 0.8], yscale='log')
    ax.set_xlabel(r'$\theta_B$ [deg]', size=12)
    ax.set_ylabel('Integrated reflectivity [rad]', size=12)
    ax.set_title(f"max miller index = {miller_max}", size=12, fontweight='bold')
    ax.grid('on')
    
    # plotting
    dcolor = {lamb0: None for lamb0 in target_lamb}
    for lamb0 in target_lamb:
        
        for ii, mat in enumerate(materials):
            
            li, = ax.plot(
                dout['reflectivity'][mat][lamb0]['Bragg'].flatten(),
                dout['reflectivity'][mat][lamb0]['Rint'].flatten(),
                ls='None',
                label=mat,
                color=dcolor[lamb0],
                **dmarker[mat],
            )
            if ii == 0:
                dcolor[lamb0] = li.get_color()
    
    # -----------------------
    # legend proxys
    
    # legend proxys for lamb
    lh = [
        mlines.Line2D([], [], color=dcolor[lamb0], ls='-')
        for lamb0 in target_lamb
    ]
    llab = [f"{lamb0*1e10: 5.3} AA" for lamb0 in target_lamb]
    legend0 = ax.legend(
        lh,
        llab,
        bbox_to_anchor=(1.05, 1),
        loc='upper left', 
        borderaxespad=0.,
    )
    ax.add_artist(legend0)
    
    # legend proxys for material
    lh = [
        mlines.Line2D([], [], color='k', ls='None', **dmarker[mat])
        for mat in materials
    ]
    ax.legend(
        lh,
        materials,
        bbox_to_anchor=(1.05, 0),
        loc='lower left', 
        borderaxespad=0.,
    )
    
    # -------------------
    # return
    # -------------------

    return dout, ax



# -*- coding: utf-8 -*-
"""
This module contains tests for tofu.geom in its structured version
"""


# Built-in
import sys
import os
import copy


# Standard
import numpy as np
import matplotlib.pyplot as plt


# tofu-specific
_PATH_HERE = os.path.abspath(os.path.dirname(__file__))
_PATH_TOFU = os.path.dirname(os.path.dirname(os.path.dirname(_PATH_HERE)))


sys.path.insert(0, _PATH_TOFU)
import tofu as tf
sys.path.pop(0)


#######################################################
#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    pass


def teardown_module():
    pass


#######################################################
#######################################################
#
#     Utilities routines
#
#######################################################


def _ref_line():

    start = np.r_[4, 0, 0]

    vect = np.r_[-1, 0, 0]
    vect = vect / np.linalg.norm(vect)

    v0 = np.r_[-vect[1], vect[0], 0.]
    v0 = v0 / np.linalg.norm(v0)

    v1 = np.cross(vect, v0)

    return start, vect, v0, v1


def _nine0e1_from_orientations(
    vect=None,
    v0=None,
    v1=None,
    theta=None,
    phi=None,
):

    nin = (
        vect * np.cos(theta)
        + np.sin(theta) * (np.cos(phi) * v0 + np.sin(phi) * v1)
    )

    e0 = (
        - vect * np.sin(theta)
        + np.cos(theta) * (np.cos(phi) * v0 + np.sin(phi) * v1)
    )

    e1 = np.cross(nin, e0)

    return nin, e0, e1


# #######################################
#              Apertures
# #######################################


def _apertures():

    start, vect, v0, v1 = _ref_line()

    # -------------------------
    # ap0 : planar from outline
    # -------------------------

    out0 = 0.01 * np.r_[-1, 1, 1, -1]
    out1 = 0.005 * np.r_[-1, -1, 1, 1]
    cent = start + 0.15 * vect
    nin, e0, e1 = _nine0e1_from_orientations(
        vect=vect,
        v0=v0,
        v1=v1,
        theta=-np.pi/20.,
        phi=0.,
    )

    ap0 = {
        'outline_x0': out0,
        'outline_x1': out1,
        'cent': cent + np.r_[0., 0.02, 0.],
        'nin': nin,
        'e0': e0,
        'e1': e1,
    }

    # ---------------------
    # ap1 : planar from 3d
    # ---------------------

    out0 = 0.05 * np.r_[-1, 1, 0]
    out1 = 0.05 * np.r_[-0.5, -0.5, 0.5]
    cent = start + 0.20 * vect
    nin, e0, e1 = _nine0e1_from_orientations(
        vect=vect,
        v0=v0,
        v1=v1,
        theta=-np.pi/10.,
        phi=np.pi/20.,
    )

    px = cent[0] + out0 * e0[0] + out1 * e1[0]
    py = cent[1] + out0 * e0[1] + out1 * e1[1]
    pz = cent[2] + out0 * e0[2] + out1 * e1[2]

    ap1 = {
        'poly_x': px,
        'poly_y': py,
        'poly_z': pz,
        'nin': nin,
    }

    # ---------------------
    # ap2 : non-planar
    # ---------------------

    out0 = 0.01 * np.r_[-1, 1, 2, 0, -2]
    out1 = 0.01 * np.r_[-1, -1, 0, 1, 0]
    cent = start + 0.3 * vect
    nin, e0, e1 = _nine0e1_from_orientations(
        vect=vect,
        v0=v0,
        v1=v1,
        theta=0.,
        phi=-np.pi/20.,
    )

    scatter = 0.001 * np.r_[-1, 1, 1, -1, -1]
    px = cent[0] + out0 * e0[0] + out1 * e1[0] + scatter * nin[0]
    py = cent[1] + out0 * e0[1] + out1 * e1[1] + scatter * nin[1]
    pz = cent[2] + out0 * e0[2] + out1 * e1[2] + scatter * nin[2]

    ap2 = {
        'poly_x': px,
        'poly_y': py,
        'poly_z': pz,
        'nin': nin,
    }

    # ----------------------------------
    # lap : non-planar for collimator
    # ----------------------------------

    out0 = 0.005 * np.r_[-1, 1, 1, -1]
    out1 = 0.003 * np.r_[-1, -1, 1, 1]

    nin, e0, e1 = _nine0e1_from_orientations(
        vect=vect,
        v0=v0,
        v1=v1,
        theta=0.,
        phi=0.,
    )

    npix = 10
    nins = np.full((npix, 3), np.nan)
    e0s = np.full((npix, 3), np.nan)
    e1s = np.full((npix, 3), np.nan)
    for ii, tt in enumerate(np.pi/20 * np.linspace(-1, 1, npix)):
        nins[ii, :], e0s[ii, :], e1s[ii, :] = _nine0e1_from_orientations(
            vect=nin,
            v0=e0,
            v1=e1,
            theta=tt,
            phi=0.,
        )

    # kl0 = 0.02 * np.linspace(-1, 1, npix)
    klin = 0.2
    delta = 0.05
    cents_x = cent[0] - (klin - delta) * nins[:, 0]    # + kl0 * e0[0]
    cents_y = cent[1] - (klin - delta) * nins[:, 1]    # + kl0 * e0[1]
    cents_z = cent[2] - (klin - delta) * nins[:, 2]    # + kl0 * e0[2]

    dap3 = {
        f'lap{ii}': {
            'outline_x0': out0,
            'outline_x1': out1,
            'cent': np.r_[cents_x[ii], cents_y[ii], cents_z[ii]],
            'nin': nins[ii, :],
            'e0': e0s[ii, :],
            'e1': e1s[ii, :],
        }
        for ii in range(npix)
    }

    # ------------
    # dout
    # -------------

    dout = {'ap0': ap0, 'ap1': ap1, 'ap2': ap2}
    dout.update(dap3)

    return dout


# #######################################
#              Filters
# #######################################


def _filters():

    dap = _apertures()

    # energy
    energ = np.linspace(1000, 5000, 100)
    trans = np.r_[
        np.zeros((10,)),
        np.linspace(0, 1, 80),
        np.ones((10,)),
    ]

    return {
        'filt0': {
            'dgeom': dict(dap['ap1']),
            'dmat': {
                'name': 'blabla',
                'symbol': 'bla',
                'thickness': 500e-6,
                'energy': energ,
                'transmission': trans,
            },
        },
    }


# #######################################
#              Cameras
# #######################################


def _cameras():

    start, vect, v0, v1 = _ref_line()

    # ---------------------
    # c0: 1d non-parallel
    # ---------------------

    out0 = 0.002 * np.r_[-1, 1, 1, -1]
    out1 = 0.005 * np.r_[-1, -1, 1, 1]
    cent = start + 0.01 * vect

    nin, e0, e1 = _nine0e1_from_orientations(
        vect=vect,
        v0=v0,
        v1=v1,
        theta=0.,
        phi=0.,
    )

    npix = 10
    nins = np.full((npix, 3), np.nan)
    e0s = np.full((npix, 3), np.nan)
    e1s = np.full((npix, 3), np.nan)
    for ii, tt in enumerate(np.pi/20 * np.linspace(-1, 1, 10)):
        nins[ii, :], e0s[ii, :], e1s[ii, :] = _nine0e1_from_orientations(
            vect=nin,
            v0=e0,
            v1=e1,
            theta=tt,
            phi=0.,
        )

    # kl0 = 0.02*np.linspace(-1, 1, 10)
    # klin = 0.01 * np.linspace(-1, 1, 10)**2
    klin = 0.2
    cents_x = cent[0] - klin * nins[:, 0]    # + kl0 * e0[0]
    cents_y = cent[1] - klin * nins[:, 1]    # + kl0 * e0[1]
    cents_z = cent[2] - klin * nins[:, 2]    # + kl0 * e0[2]

    c0 = {
        'dgeom': {
            'outline_x0': out0,
            'outline_x1': out1,
            'cents_x': cents_x,
            'cents_y': cents_y,
            'cents_z': cents_z,
            'nin_x': nins[:, 0],
            'nin_y': nins[:, 1],
            'nin_z': nins[:, 2],
            'e0_x': e0s[:, 0],
            'e0_y': e0s[:, 1],
            'e0_z': e0s[:, 2],
            'e1_x': e1s[:, 0],
            'e1_y': e1s[:, 1],
            'e1_z': e1s[:, 2],
        },
    }

    # -------------------------
    # c1: 1d parallel coplanar
    # -------------------------

    kl = 0.02*np.linspace(-1, 1, 10)
    cents_x = cent[0] + kl * e0[0]
    cents_y = cent[1] + kl * e0[1]
    cents_z = cent[2] + kl * e0[2]

    c1 = {
        'dgeom': {
            'outline_x0': out0,
            'outline_x1': out1,
            'cents_x': cents_x,
            'cents_y': cents_y,
            'cents_z': cents_z,
            'nin_x': nins[0, 0],
            'nin_y': nins[0, 1],
            'nin_z': nins[0, 2],
            'e0_x': e0s[0, 0],
            'e0_y': e0s[0, 1],
            'e0_z': e0s[0, 2],
            'e1_x': e1s[0, 0],
            'e1_y': e1s[0, 1],
            'e1_z': e1s[0, 2],
        },
        'dmat': {
            'qeff_E': np.linspace(1, 10, 100)*1e3,
            'qeff': 0.99*np.ones((100,)),
        },
    }

    # ---------------------
    # c2: 2d
    # ---------------------

    out0 = 0.005 * np.r_[-1, 1, 1, -1]
    out1 = 0.005 * np.r_[-1, -1, 1, 1]
    cent = start + 0.01 * vect

    cent0 = 0.02*np.linspace(-1, 1, 5)
    cent1 = 0.02*np.linspace(-1, 1, 5)

    nin, e0, e1 = _nine0e1_from_orientations(
        vect=vect,
        v0=v0,
        v1=v1,
        theta=-np.pi/20,
        phi=0.,
    )

    c2 = {
        'dgeom': {
            'outline_x0': out0,
            'outline_x1': out1,
            'cent': cent,
            'cents_x0': cent0,
            'cents_x1': cent1,
            'nin': nin,
            'e0': e0,
            'e1': e1,
        },
    }

    return {
        'cam0': c0,
        'cam1': c1,
        'cam2': c2,
        'cam00': copy.deepcopy(c0),
        'cam11': copy.deepcopy(c1),
        'cam22': copy.deepcopy(c2),
        'cam000': copy.deepcopy(c0),
        'cam111': copy.deepcopy(c1),
        'cam222': copy.deepcopy(c2),
        'cam0000': copy.deepcopy(c0),
        'cam00000': copy.deepcopy(c0),
    }


# #######################################
#              Crystals
# #######################################


def _crystals():

    start, vect, v0, v1 = _ref_line()

    # cryst0: planar
    cent = start + 0.01 * vect

    nin, e0, e1 = _nine0e1_from_orientations(
        vect=vect,
        v0=v0,
        v1=v1,
        theta=-50*np.pi/180,
        phi=0.,
    )

    size = 2.e-2
    c0 = {
        'dgeom': {
            'cent': cent,
            'nin': nin,
            'e0': e0,
            'e1': e1,
            'extenthalf': size,
            'curve_r': np.inf,
        },
        'dmat': 'Quartz_110',
    }

    # c1: cylindrical
    size = 6.e-2
    rc = 2.
    c1 = {
        'dgeom': {
            'cent': cent,
            'nin': nin,
            'e0': e0,
            'e1': e1,
            'extenthalf': size * np.r_[1, 1/rc],
            'curve_r': [np.inf, rc],
        },
        'dmat': 'Quartz_110',
    }

    # c2: spherical
    rc = 2.
    c2 = {
        'dgeom': {
            'cent': cent,
            'nin': nin,
            'e0': e0,
            'e1': e1,
            'extenthalf': size * np.r_[1 / rc, 1 / rc],
            'curve_r': rc,
        },
        'dmat': 'Quartz_110',
    }

    # c3: toroidall
    rc = 2.
    c3 = {
        'dgeom': {
            'cent': cent,
            'nin': nin,
            'e0': e0,
            'e1': e1,
            'extenthalf': size * np.r_[1 / (3.*rc/2.), 1 / (rc/2.)],
            'curve_r': [rc, rc/2.],
        },
        'dmat': 'Quartz_110',
    }

    return {
        'cryst0': c0,
        'cryst1': c1,
        'cryst2': c2,
        'cryst3': c3,
    }


def _configurations():
    return {
        'cryst0': ['pinhole'],
        'cryst1': ['pinhole', 'von hamos'],
        'cryst2': ['pinhole', 'johann'],
        'cryst3': [],
    }


# #######################################
#              Diagnostics
# #######################################


def _diagnostics():

    # d0: single 1d camera
    d0 = {'doptics': 'cam0'}

    # d1: single 2d camera
    d1 = {'doptics': 'cam1', 'resolution': 0.2}

    # d2: single 2d camera
    d2 = {'doptics': 'cam2'}

    # d3: 1d + 1 aperture
    d3 = {'doptics': ['cam00', 'ap0']}

    # d4: 2d + 1 aperture
    d4 = {'doptics': ['cam11', 'ap0']}

    # d5: 2d + 1 aperture
    d5 = {'doptics': ['cam22', 'ap0']}

    # d6: 1d + multiple apertures
    d6 = {'doptics': ['cam000', 'ap0', 'filt0', 'ap2']}

    # d7: 1d parallel coplanar + multiple apertures
    d7 = {'doptics': ['cam111', 'ap0', 'filt0', 'ap2']}

    # d8: 2d + multiple apertures
    d8 = {'doptics': ['cam222', 'ap0', 'filt0', 'ap2']}

    # # d9: 2d + spherical crystal
    # d9 = {'optics': ('c3','cryst0')}

    # # d10: 2d + cylindrical crystal + slit
    # d10 = {'optics': ('c3','cryst1', 'slit0')}

    # # d11: 2d + toroidal crystal + slit
    # d11 = {'optics': ('c3','cryst2', 'slit1')}

    # d12: 1d collimator camera, one aperture per pixel
    d12 = {
        'doptics': tuple(
            ['cam0000'] + [k1 for k1 in [f'lap{ii}' for ii in range(10)]]
        ),
    }

    # d13: 1d collimator-hybrid camera
    d13 = {
        'doptics': {
            'cam00000': {
                'optics': [f'lap{ii}' for ii in range(10)] + ['ap0'],
                'paths': np.concatenate(
                    (np.eye(10), np.ones((10, 1))),
                    axis=1,
                ),
            },
        },
    }

    return {
        'diag0': d0,
        'diag1': d1,
        'diag2': d2,
        'diag3': d3,
        'diag4': d4,
        'diag5': d5,
        'diag6': d6,
        'diag7': d7,
        'diag8': d8,
        # 'd9': d9,
        # 'd10': d10,
        # 'd11': d11,
        'd12': d12,
        'd13': d13,
    }


#######################################################
#######################################################
#
#     Tests
#
#######################################################


class Test01_Diagnostic():

    def setup_method(self):

        # get config
        conf = tf.load_config('SPARC')
        conf.remove_Struct(Cls='PFC', Name='ICRH0')
        self.conf = conf

        # IRCH0 for get_touch, shifted
        conf_touch = tf.load_config('SPARC')
        poly = conf_touch.PFC.ICRH0.Poly
        poly[0, :] = poly[0, :] - 0.05
        conf_touch.remove_Struct(Cls='PFC', Name='ICRH0')
        conf_touch.add_Struct(
            Cls='PFC',
            Name='ICRH0',
            Poly=poly,
            Lim=np.r_[20, 340]*np.pi/180.,
            dextraprop={'visible': True},
        )
        self.conf_touch = conf_touch

        # get dict
        dapertures = _apertures()
        dfilters = _filters()
        dcameras = _cameras()
        dcrystals = _crystals()
        dconfig = _configurations()
        ddiag = _diagnostics()

        # instanciate
        self.coll = tf.data.Collection()

        # add apertures
        for k0, v0 in dapertures.items():
            self.coll.add_aperture(key=k0, **v0)

        # add filters
        for k0, v0 in dfilters.items():
            self.coll.add_filter(key=k0, **v0)

        # add cameras
        for k0, v0 in dcameras.items():
            if 'cam0' in k0 or 'cam1' in k0:
                self.coll.add_camera_1d(key=k0, **v0)
            else:
                self.coll.add_camera_2d(key=k0, **v0)

        # add diagnostics
        for k0, v0 in ddiag.items():
            self.coll.add_diagnostic(
                key=k0,
                config=conf,
                reflections_nb=1,
                reflections_type='specular',
                **v0,
            )

        # add crystals
        for k0, v0 in dcrystals.items():
            self.coll.add_crystal(key=k0, **v0)

        # add crystal optics
        self.doptics = {}
        for k0, v0 in dcrystals.items():

            for ii, cc in enumerate(dconfig[k0]):

                apdim = [100e-6, 8e-2] if cc != 'pinhole' else None
                pinrad = 1e-2 if cc == 'pinhole' else None

                loptics = self.coll.get_crystal_ideal_configuration(
                    key=k0,
                    configuration=cc,
                    # parameters
                    cam_on_e0=False,
                    cam_tangential=True,
                    cam_dimensions=[8e-2, 5e-2],
                    focal_distance=2.,
                    # store
                    store=True,
                    key_cam=f'{k0}_cam{ii}',
                    aperture_dimensions=apdim,
                    pinhole_radius=pinrad,
                    cam_pixels_nb=[5, 3],
                    # returnas
                    returnas=list,
                )

                if 'cryst1_slit' in loptics:
                    loptics.append('ap0')

                # add diag
                gtype = self.coll.dobj['crystal'][k0]['dgeom']['type']
                if gtype not in ['toroidal']:
                    self.doptics.update({
                        f'{k0}_{ii}': loptics,
                    })

        # add crystal optics
        for k0, v0 in self.doptics.items():
            self.coll.add_diagnostic(
                doptics=v0,
                config=self.conf,
                compute_vos_from_los=True,
            )
        # add toroidal
        # self.coll.add_diagnostic(optics=['cryst2-cam0', 'cryst3'])

    # ----------
    # tests

    """
    def test01_etendues(self, res=np.r_[0.005, 0.003, 0.001]):
        for k0, v0 in self.coll.dobj['diagnostic'].items():
            if len(v0['optics']) == 1 or v0['spectro'] is not False:
                continue
            self.coll.compute_diagnostic_etendue_los(
                key=k0,
                res=res,
                numerical=True,
                analytical=True,
                check=True,
                store=False,
            )
            plt.close('all')
    """

    def test02_get_outline(self):

        # apertures
        for k0, v0 in self.coll.dobj['aperture'].items():
            _ = self.coll.get_optics_outline(k0)

        # camera
        for k0, v0 in self.coll.dobj['camera'].items():
            _ = self.coll.get_optics_outline(k0)

        # crystals
        for k0, v0 in self.coll.dobj['crystal'].items():
            _ = self.coll.get_optics_outline(k0)

    def test03_plot(self):
        for ii, (k0, v0) in enumerate(self.coll.dobj['diagnostic'].items()):
            dax = self.coll.plot_diagnostic(
                k0,
                data='etendue',
                proj=(
                    None if ii % 3 == 0
                    else ('cross' if ii % 3 == 1 else ['cross', 'hor'])
                ),
            )
            plt.close('all')
            del dax

    def test04_sinogram(self):

        lrays = list(self.coll.dobj['rays'].keys())
        ldiag = [
            k0 for k0, v0 in self.coll.dobj['diagnostic'].items()
            if any([
                v1.get('los') is not None
                for v1 in v0['doptics'].values()
            ])
        ]
        lk = [lrays] + ldiag
        for ii, k0 in enumerate(lk):

            dout, dax = self.coll.get_sinogram(
                key=k0,
                ang='theta' if ii % 2 == 0 else 'xi',
                ang_units='deg' if ii % 3 == 0 else 'radian',
                impact_pos=ii % 3 != 0,
                R0=2.4 if ii % 3 != 1 else None,
                config=None if ii % 3 != 1 else self.conf,
                pmax=None if ii % 3 == 0 else 5,
                plot=True,
                verb=2,
            )
            plt.close('all')
            del dax

    def test05_vos(self):

        # add mesh
        key_mesh = 'm0'

        if key_mesh not in self.coll.dobj.get('mesh', {}).keys():
            self.coll.add_mesh_2d_rect(
                key=key_mesh,
                res=0.1,
                crop_poly=self.conf,
            )

        # compute vos
        for ii, (k0, v0) in enumerate(self.coll.dobj['diagnostic'].items()):

            doptics = self.coll.dobj['diagnostic'][k0]['doptics']
            lcam = self.coll.dobj['diagnostic'][k0]['camera']
            if len(doptics[lcam[0]]['optics']) == 0:
                continue

            self.coll.compute_diagnostic_vos(
                # keys
                key_diag=k0,
                key_mesh=key_mesh,
                # resolution
                res_RZ=0.04,
                res_phi=0.04,
                # spectro
                n0=3,
                n1=3,
                res_lamb=2e-10,
                visibility=False,
                store=True,
            )

    def test06_plot_coverage(self):

        # add mesh
        key_mesh = 'm0'

        if key_mesh not in self.coll.dobj.get('mesh', {}).keys():
            self.coll.add_mesh_2d_rect(
                key=key_mesh,
                res=0.1,
                crop_poly=self.conf,
            )

        for ii, (k0, v0) in enumerate(self.coll.dobj['diagnostic'].items()):
            lcam = self.coll.dobj['diagnostic'][k0]['camera']
            doptics = self.coll.dobj['diagnostic'][k0]['doptics']
            if len(doptics[lcam[0]]['optics']) == 0:
                continue
            if self.coll.dobj['diagnostic'][k0]['spectro']:
                continue

            if doptics[lcam[0]]['dvos'].get('keym') is None:
                self.coll.compute_diagnostic_vos(
                    # keys
                    key_diag=k0,
                    key_mesh=key_mesh,
                    # resolution
                    res_RZ=0.03,
                    res_phi=0.04,
                    # spectro
                    n0=5,
                    n1=5,
                    res_lamb=1e-10,
                    visibility=False,
                    store=True,
                )
            _ = self.coll.plot_diagnostic_geometrical_coverage(k0)

        plt.close('all')

    def test07_add_single_point_camera2d(self):

        self.coll.add_single_point_camera2d(
            key='ptcam',
            key_rays='diag5_cam22_los',
            angle0=55,
            angle1=55,
            config=self.conf_touch,
        )

        # add rays
        dsamp = {'dedge': {'res': 'min'}, 'dsurface': {'nb': 3}}
        self.coll.add_rays_from_diagnostic(
            key='diag5',
            dsampling_pixel=dsamp,
            dsampling_optics=dsamp,
            optics=-1,
            config=self.conf_touch,
            store=True,
            strict=None,
            key_rays='diag5_cam22_rays',
            overwrite=None,
        )

        # get angles from rays
        dout = self.coll.get_rays_angles_from_single_point_camera2d(
            key_single_pt_cam='ptcam',
            key_rays='diag5_cam22_los',
            return_indices=False,
        )

        # get angles from rays
        dout = self.coll.get_rays_angles_from_single_point_camera2d(
            key_single_pt_cam='ptcam',
            key_rays='diag5_cam22_rays',
            return_indices=True,
            convex_axis=False,
        )

        # get angles from rays
        dout = self.coll.get_rays_angles_from_single_point_camera2d(
            key_single_pt_cam='ptcam',
            key_rays='diag5_cam22_rays',
            return_indices=True,
            convex_axis=(-1, -2),
        )
        assert isinstance(dout, dict)

    def test07_add_rays_from_diagnostic(self):
        for ii, (k0, v0) in enumerate(self.coll.dobj['diagnostic'].items()):
            noptics = any([
                len(v1['optics']) == 0 for v1 in v0['doptics'].values()
            ])
            if v0['is2d'] or v0['spectro'] or noptics:
                continue
            dsamp = {'dedge': {'res': 'max'}, 'dsurface': {'nb': 3}}
            dout = self.coll.add_rays_from_diagnostic(
                key=k0,
                dsampling_pixel=dsamp,
                dsampling_optics=dsamp,
                optics=-1,
                config=self.conf,
                store=(ii % 2 == 0),
            )
            assert isinstance(dout, dict) or dout is None

    def test08_reverse_ray_tracing(self):
        for ii, (k0, v0) in enumerate(self.coll.dobj['diagnostic'].items()):
            lcam = self.coll.dobj['diagnostic'][k0]['camera']
            doptics = self.coll.dobj['diagnostic'][k0]['doptics']
            if len(doptics[lcam[0]]['optics']) == 0:
                continue
            if not self.coll.dobj['diagnostic'][k0]['spectro']:
                continue

            # Get points
            ptsx, ptsy, ptsz = self.coll.get_rays_pts(k0)
            kcryst = list(doptics.values())[0]['optics'][0]
            lamb = self.coll.get_crystal_bragglamb(
                key=kcryst,
                lamb=None,
                bragg=None,
                norder=None,
                rocking_curve=None,
            )[1]

            _ = self.coll.get_raytracing_from_pts(
                key=k0,
                key_cam=None,
                key_mesh=None,
                res_RZ=None,
                res_phi=None,
                ptsx=ptsx[-1, ...],
                ptsy=ptsy[-1, ...],
                ptsz=ptsz[-1, ...],
                n0=3,
                n1=3,
                lamb0=lamb,
                res_lamb=None,
                rocking_curve=None,
                res_rock_curve=None,
                append=None,
                plot=True,
                dax=None,
                plot_pixels=None,
                plot_config=None,
                vmin=None,
                vmax=None,
                aspect3d=None,
                elements=None,
                colorbar=None,
            )

    def test09_get_rays_touch_dict(self):
        dout = self.coll.get_rays_touch_dict(
            key='diag5_cam22_los',
            config=self.conf_touch,
            segment=-1,
            allowed=['PFC_ICRH0', 'Ves_FirstWallV0'],
        )
        assert isinstance(dout, dict)

    def test10_save_to_json(self):

        for ii, (k0, v0) in enumerate(self.coll.dobj['diagnostic'].items()):

            lcam = self.coll.dobj['diagnostic'][k0]['camera']
            doptics = self.coll.dobj['diagnostic'][k0]['doptics']
            if len(doptics[lcam[0]]['optics']) == 0 or k0 == 'diag6':
                continue

            # saving
            pfe = os.path.join(_PATH_HERE, f"{k0}.json")
            self.coll.save_diagnostic_to_file(k0, pfe_save=pfe)

            # reloading
            _ = tf.data.load_diagnostic_from_file(pfe)

            # remove file
            os.remove(pfe)





###########################################################
###########################################################
#       Define default Experiment names
###########################################################


_ExpWest = 'WEST'
_ExpJET = 'JET'
_ExpITER = 'ITER'
_ExpAUG = 'AUG'
_ExpDEMO = 'DEMO'
_ExpTOMAS = 'TOMAS'
_ExpCOMPASS = 'COMPASS'
_ExpCOMPASS2 = 'COMPASS2'
_ExpTCV = 'TCV'
_ExpSPARC = 'SPARC'
_ExpNSTX = 'NSTX'
_ExpKSTAR = 'KSTAR'
_ExpMAST = 'MAST'


# Default config
_DEFCONFIG = _ExpITER
# coils taken from:
# ids=['wall', 'pf_active'],
# shot=1180, run=17, database='ITER_MD', user='public'


###########################################################
###########################################################
#       Define default Configs
###########################################################


# Dictionnary of unique config names
# For each config, indicates which structural elements it comprises
# Elements are sorted by class (Ves, PFC...)
# For each element, a unique txt file containing the geometry will be loaded
_DCONFIG = {

    # WEST
    'WEST-V0': {
        'Exp': _ExpWest,
        'Ves': ['V1'],
        'CoilPF': [
            'CSV0',
            'BlV0', 'BuV0',
            'DivLow1V0', 'DivLow2V0',
            'DivUp1V0', 'DivUp2V0',
            'DlV0', 'DuV0', 'ElV0', 'EuV0', 'FlV0', 'FuV0',
        ],
    },
    'WEST-Sep': {
        'Exp': _ExpWest,
        'PlasmaDomain': ['Sep'],
        'CoilPF': [
            'CSV0',
            'BlV0', 'BuV0',
            'DivLow1V0', 'DivLow2V0',
            'DivUp1V0', 'DivUp2V0',
            'DlV0', 'DuV0', 'ElV0', 'EuV0', 'FlV0', 'FuV0',
        ],
    },
    'WEST-V1': {
        'Exp': _ExpWest,
        'Ves': ['V2'],
        'PFC': ['BaffleV0', 'DivUpV1', 'DivLowITERV1'],
        'CoilPF': [
            'CSV0',
            'BlV0', 'BuV0',
            'DivLow1V0', 'DivLow2V0',
            'DivUp1V0', 'DivUp2V0',
            'DlV0', 'DuV0', 'ElV0', 'EuV0', 'FlV0', 'FuV0',
        ],
    },
    'WEST-V2': {
        'Exp': _ExpWest,
        'Ves': ['V2'],
        'PFC': [
            'BaffleV1', 'DivUpV2', 'DivLowITERV2',
            'BumperInnerV1', 'BumperOuterV1',
            'IC1V1', 'IC2V1', 'IC3V1',
        ],
        'CoilPF': [
            'CSV0',
            'BlV0', 'BuV0',
            'DivLow1V0', 'DivLow2V0',
            'DivUp1V0', 'DivUp2V0',
            'DlV0', 'DuV0', 'ElV0', 'EuV0', 'FlV0', 'FuV0',
        ],
    },
    'WEST-V3': {
        'Exp': _ExpWest,
        'Ves': ['V2'],
        'PFC': [
            'BaffleV2', 'DivUpV3', 'DivLowITERV3',
            'BumperInnerV3', 'BumperOuterV3',
            'IC1V1', 'IC2V1', 'IC3V1',
            'LH1V1', 'LH2V1',
            'RippleV1', 'VDEV0',
        ],
        'CoilPF': [
            'CSV0',
            'BlV0', 'BuV0',
            'DivLow1V0', 'DivLow2V0',
            'DivUp1V0', 'DivUp2V0',
            'DlV0', 'DuV0', 'ElV0', 'EuV0', 'FlV0', 'FuV0',
        ],
    },
    'WEST-V4': {
        'Exp': _ExpWest,
        'Ves': ['InnerV0', 'OuterV0'],
        'PFC': [
            'BaffleV2', 'DivUpV3', 'DivLowITERV3',
            'BumperInnerV3', 'BumperOuterV3',
            'RippleV1', 'VDEV0',
            'ThermalShieldHFSV0',
            'ThermalShieldLFSSlimV0',
            'ThermalShieldLFSWideV0',
            'ThermalShieldLFSLowV0',
            'ThermalShieldLFSUpV0',
            'LH1V1', 'LH2V1',
            'IC1V1', 'IC2V1', 'IC3V1',
            'CasingCoverLDivV0', 'CasingCoverUDivV0',
            'CasingLDivV0', 'CasingUDivV0',
            'CasingPFUPlateLDivV0', 'CasingPFUPlateUDivV0',
            'CasingPJLDivV0', 'CasingPJUDivV0',
        ],
        'CoilPF': [
            'CSV0',
            'BlV0', 'BuV0',
            'DivLow1V0', 'DivLow2V0',
            'DivUp1V0', 'DivUp2V0',
            'DlV0', 'DuV0', 'ElV0', 'EuV0', 'FlV0', 'FuV0',
        ],
    },

    # JET
    'JET-V0': {
        'Exp': _ExpJET,
        'Ves': ['V0'],
    },

    # ITER
    'ITER-V0': {
        'Exp': _ExpITER,
        'Ves': ['V0'],
        'CoilPF': ['PF1', 'PF2', 'PF3', 'PF4', 'PF5', 'PF6'],
        'CoilCS': ['CS3U', 'CS2U', 'CS1U', 'CS1L', 'CS2L', 'CS3L'],
    },
    'ITER-V1': {
        'Exp': _ExpITER,
        'Ves': ['V1'],
        'PFC': [
            'BLK01', 'BLK02', 'BLK03', 'BLK04', 'BLK05', 'BLK06',
            'BLK07', 'BLK08', 'BLK09', 'BLK10', 'BLK11', 'BLK12',
            'BLK13', 'BLK14', 'BLK15', 'BLK16', 'BLK17', 'BLK18',
            'Div1', 'Div2', 'Div3', 'Div4', 'Div5', 'Div6',
        ],
        'CoilPF': ['PF1', 'PF2', 'PF3', 'PF4', 'PF5', 'PF6'],
        'CoilCS': ['CS3U', 'CS2U', 'CS1U', 'CS1L', 'CS2L', 'CS3L'],
    },
    'ITER-V2': {
        'Exp': _ExpITER,
        'Ves': ['InnerV0', 'OuterV0', 'Cryostat'],
        'PFC': [
            'BLK01', 'BLK02', 'BLK03', 'BLK04', 'BLK05', 'BLK06',
            'BLK07', 'BLK08', 'BLK09', 'BLK10', 'BLK11', 'BLK12',
            'BLK13', 'BLK14', 'BLK15', 'BLK16', 'BLK17', 'BLK18',
            'Div1', 'Div2', 'Div3', 'Div4', 'Div5', 'Div6',
        ],
        'CoilPF': ['PF1', 'PF2', 'PF3', 'PF4', 'PF5', 'PF6'],
        'CoilCS': ['CS3U', 'CS2U', 'CS1U', 'CS1L', 'CS2L', 'CS3L'],
    },
    # 'ITER-SOLEDGE3XV0': {
    # 'Exp': _ExpITER,
    # 'Ves': ['SOLEDGE3XV0'],
    # 'PFC': ['SOLEDGE3XDivDomeV0', 'SOLEDGE3XDivSupportV0']
    # },

    # AUG
    'AUG-V0': {
        'Exp': _ExpAUG,
        'Ves': ['V0'],
    },
    'AUG-V1': {
        'Exp': _ExpAUG,
        'Ves': ['VESiR'],
        'PFC': [
            'D2cdome', 'D2cdomL', 'D2cdomR', 'D2ci1',
            'D2ci2', 'D2cTPib', 'D2cTPic', 'D2cTPi',
            'D2dBG2', 'D2dBl1', 'D2dBl2', 'D2dBl3',
            'D2dBu1', 'D2dBu2', 'D2dBu3', 'D2dBu4',
            'D3BG10', 'D3BG1', 'ICRHa', 'LIM09', 'PClow',
            'PCup', 'SBi', 'TPLT1', 'TPLT2', 'TPLT3',
            'TPLT4', 'TPLT5', 'TPRT2', 'TPRT3', 'TPRT4',
            'TPRT5',
        ],
    },

    # DEMO
    'DEMO-2019': {
        'Exp': _ExpDEMO,
        'Ves': ['V0'],
        'PFC': [
            'LimiterUpperV0', 'LimiterEquatV0',
            'BlanketInnerV0', 'BlanketOuterV0', 'DivertorV0'],
    },

    # TOMAS
    'TOMAS-V0': {
        'Exp': _ExpTOMAS,
        'Ves': ['V0'],
        'PFC': ['LimiterV0', 'AntennaV0'],
    },

    # COMPASS
    'COMPASS-V0': {
        'Exp': _ExpCOMPASS,
        'Ves': ['V0'],
    },
    'COMPASS-V1': {
        'Exp': _ExpCOMPASS,
        'Ves': ['InnerV1'],
        'PFC': ['lower', 'upper', 'inner', 'outer'],
    },

    # COMPASS2
    'COMPASS2-V0': {
        'Exp': _ExpCOMPASS2,
        'Ves': ['V0'],
    },

    # TCV
    'TCV-V0': {
        'Exp': _ExpTCV,
        'Ves': ['vIn', 'vOut', 't'],
        'CoilPF': [
            'A001', 'B001', 'B002',
            'B03A1', 'B03A2', 'B03A3',
            'C001', 'C002',
            'D001', 'D002',
            'E001', 'E002', 'E003', 'E004', 'E005', 'E006', 'E007', 'E008',
            'E03A1', 'E03A2', 'E03A3',
            'F001', 'F002', 'F003', 'F004', 'F005', 'F006', 'F007', 'F008',
            'T03A1', 'T03A2', 'T03A3'],
    },

    # SPARC
    'SPARC-V0': {
        'Exp': _ExpSPARC,
        'Ves': ['FirstWallV0'],
    },

    'SPARC-V1': {
        'Exp': _ExpSPARC,
        'Ves': ['FirstWallV0', 'VesInner', 'VesOuter'],
        'PFC': ['ICRH0'],
        'CoilPF': [
            'Div1lower', 'Div2lower', 'Div1upper', 'Div2upper',
            'EFClower0', 'EFClower1', 'EFCmed0', 'EFCmed1',
            'EFCupper0', 'EFCupper1', 'PF1lower', 'PF1upper',
            'PF2lower', 'PF2upper', 'PF3lower', 'PF3upper',
            'PF4lower', 'PF4upper', 'VS1lower', 'VS1upper',
            'VStabPlatelower', 'VStabPlateupper',
        ],
        'CoilCS': [
            'CS1lower', 'CS2lower', 'CS3lower',
            'CS1upper', 'CS2upper', 'CS3upper',
        ],
    },

    'SPARC-V2': {
        'Exp': _ExpSPARC,
        'Ves': [
            'FirstWallV0', 'VesInner', 'VesOuter',
            'CoilTFInner', 'CoilTFOuter',
        ],
        'PFC': ['ICRH0'],
        'CoilPF': [
            'Div1lower', 'Div2lower', 'Div1upper', 'Div2upper',
            'EFClower0', 'EFClower1', 'EFCmed0', 'EFCmed1',
            'EFCupper0', 'EFCupper1', 'PF1lower', 'PF1upper',
            'PF2lower', 'PF2upper', 'PF3lower', 'PF3upper',
            'PF4lower', 'PF4upper', 'VS1lower', 'VS1upper',
            'VStabPlatelower', 'VStabPlateupper',
        ],
        'CoilCS': [
            'CS1lower', 'CS2lower', 'CS3lower',
            'CS1upper', 'CS2upper', 'CS3upper',
        ],
    },

    # NSTX
    'NSTX-V0': {
        'Exp': _ExpNSTX,
        'Ves': ['V0'],
    },
    'NSTX-V1': {
        'Exp': _ExpNSTX,
        'Ves': ['VesselInner'],
        'CoilPF': [
            'CentralSolenoid', 'PFCoil01', 'PFCoil02', 'PFCoil03',
            'PFCoil04', 'PFCoil05', 'PFCoil06', 'PFCoil07', 'PFCoil08',
            'PFCoil09', 'PFCoil10', 'PFCoil11', 'PFCoil12', 'PFCoil13',
            'PFCoil14', 'PFCoil16', 'PFCoil17', 'PFCoil18', 'PFCoil19',
            'PFCoil20', 'PFCoil21', 'PFCoil22',
        ],
    },
    'NSTX-V2': {
        'Exp': _ExpNSTX,
        'Ves': ['VesselInner'],
        'PFC': [
            'VesselOutter01', 'VesselOutter02',
            'BumperOutter01', 'BumperOutter02',
            'BumperOutter03', 'BumperOutter04',
            'DivertorUpper', 'DivertorLower',
            'ICRFAntenna',
            'path69778', 'path69912', 'path69980', 'path70015',
            'path70085', 'path70153',
        ],
        'CoilPF': [
            'CentralSolenoid', 'PFCoil01', 'PFCoil02', 'PFCoil03',
            'PFCoil04', 'PFCoil05', 'PFCoil06', 'PFCoil07', 'PFCoil08',
            'PFCoil09', 'PFCoil10', 'PFCoil11', 'PFCoil12', 'PFCoil13',
            'PFCoil14', 'PFCoil16', 'PFCoil17', 'PFCoil18', 'PFCoil19',
            'PFCoil20', 'PFCoil21', 'PFCoil22',
        ],
    },

    # KSTAR
    'KSTAR-V0': {
        'Exp': _ExpKSTAR,
        'Ves': ['V0'],
    },

    # MAST
    'MAST-V0': {
        'Exp': _ExpMAST,
        'Ves': ['V0'],
    },
}

# Each config can be called by various names / shortcuts (for benchmark and
# retro-compatibility), this table stores, for each shortcut,
# the associated unique name it refers to
_DCONFIG_SHORTCUTS = {
    'ITER': 'ITER-V2',
    # 'ITER-SOLEDGE3X': 'ITER-SOLEDGE3XV0',
    'JET': 'JET-V0',
    'WEST': 'WEST-V4',
    'A1': 'WEST-V1',
    'A2': 'ITER-V1',
    'A3': 'WEST-Sep',
    'B1': 'WEST-V2',
    'B2': 'WEST-V3',
    'B3': 'WEST-V4',
    'B4': 'ITER-V2',
    'AUG': 'AUG-V1',
    'DEMO': 'DEMO-2019',
    'TOMAS': 'TOMAS-V0',
    'COMPASS': 'COMPASS-V0',
    'TCV': 'TCV-V0',
    'SPARC': 'SPARC-V1',
    'NSTX': 'NSTX-V2',
    'KSTAR': 'KSTAR-V0',
    'MAST': 'MAST-V0',
}


# Check all shortcuts
lout = []
for k0, v0 in _DCONFIG_SHORTCUTS.items():
    if v0 not in _DCONFIG.keys():
        lout.append('{}: {}'.format(k0, v0))
if len(lout) > 0:
    msg = (
        "\nThe following shortcuts refer to undefined config names:\n"
        + "\t- "
        + "\n\t- ".join(lout)
        + "\nAvailable config names are:\n"
        + "\t- "
        + "\n\t- ".join(_DCONFIG.keys())
    )
    raise Exception(msg)





import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.lines as mlines
import datastock as ds


# ################################################
# ################################################
#           Main
# ################################################


def main(
    det_size=None,
    ap_size=None,
    focal=None,
    pitch=None,
    det_nb=None,
    dist=None,
    # equivalent collimator
    collimator=None,
    collimator_length=None,
    diverging=None,
):
    """ Plot a 2d represntation of a pinhole camera

    Simple planar pinhole camera model, with parameterized:
        - det_size: detector size
        - ap_size: aperture size
        - focal: focal length
        - pitch: distance between sensors center to center
        - det_nb: nb of sensors, has to be an odd integer
        - dist: dist up to which plotting should be done

    """

    # ---------------
    # check inputs
    # ---------------

    (
        det_size,
        ap_size,
        focal,
        pitch,
        det_nb,
        dist,
        # equivalent collimator
        collimator,
        collimator_length,
        diverging,
        kcase,
    ) = _check(
        det_size=det_size,
        ap_size=ap_size,
        focal=focal,
        pitch=pitch,
        det_nb=det_nb,
        dist=dist,
        # equivalent collimator
        collimator=collimator,
        collimator_length=collimator_length,
        diverging=diverging,
    )

    # ---------------
    # inputs
    # ---------------

    dout = {
        'inputs': {
            # det
            'det_size': det_size,
            'det_nb': det_nb,
            'pitch': pitch,
            # ap
            'ap_size': ap_size,
            'focal': focal,
            # plot
            'dist': dist,
        }
    }
    if collimator is True:
        dout['inputs']['length'] = collimator_length

    # ---------------
    # compute sensors
    # ---------------

    dout['sensors'] = _compute_sensors(**dout['inputs'])

    # ---------------
    # compute pinhole (ref)
    # ---------------

    dout['pinhole'] = _compute_pinhole(
        dinputs=dout['inputs'],
        dsensors=dout['sensors'],
    )

    # ---------------
    # compute collimator
    # ---------------

    if collimator is True:

        # pick function
        if diverging is True:
            func = _compute_collimator_diverging
        else:
            func = _compute_collimator_converging

        # compute
        dout[kcase] = func(
            dinputs=dout['inputs'],
            dsensors=dout['sensors'],
            dpinhole=dout['pinhole'],
            # equivalent collimator
            length=collimator_length,
            diverging=diverging,
        )

    # ---------------
    # add text
    # --------------

    dout['text'] = {}

    # inputs
    dout['text'].update(_add_text_inputs(dout, collimator))
    dout['text'].update(_add_text_pinhole(dout))

    if collimator is True:
        if diverging is True:
            dout['text'].update(_add_text_collimator_diverging(dout, kcase))
        else:
            dout['text'].update(_add_text_collimator_converging(dout, kcase))

    # ---------------
    # plot
    # ---------------

    _plot(
        dout=dout,
        kcase=kcase,
    )

    return dout


# ################################################
# ################################################
#           Check
# ################################################


def _check(
    det_size=None,
    ap_size=None,
    focal=None,
    pitch=None,
    det_nb=None,
    dist=None,
    # equivalent collimator
    collimator=None,
    collimator_length=None,
    diverging=None,
):

    # -------------
    # sizes
    # -------------

    # det_size
    det_size = float(ds._generic_check._check_var(
        det_size, 'det_size',
        types=(int, float),
        sign='>0',
        default=0.0013,
    ))

    # ap_size
    ap_size = float(ds._generic_check._check_var(
        ap_size, 'ap_size',
        types=(int, float),
        sign='>0',
        default=0.0043,
    ))

    # -------------
    # focal
    # -------------

    focal = float(ds._generic_check._check_var(
        focal, 'focal',
        types=(int, float),
        sign='>0',
        default=0.20,
    ))

    # -------------
    # pitch
    # -------------

    pitch = float(ds._generic_check._check_var(
        pitch, 'pitch',
        types=(int, float),
        sign='>0',
        default=0.005,
    ))

    # -------------
    # det_nb
    # -------------

    det_nb = ds._generic_check._check_var(
        det_nb, 'det_nb',
        types=int,
        sign='>0',
        default=3,
    )

    if det_nb % 2 == 0:
        det_nb = det_nb + 1

    # -------------
    # dist
    # -------------

    dist = float(ds._generic_check._check_var(
        dist, 'dist',
        types=(int, float),
        sign='>0',
        default=2.,
    ))

    # -------------
    # collimator
    # -------------

    collimator = ds._generic_check._check_var(
        collimator, 'collimator',
        types=bool,
        default=(collimator_length is not None),
    )

    # -------------
    # collimator_length
    # -------------

    collimator_length = float(ds._generic_check._check_var(
        collimator_length, 'collimator_length',
        types=(int, float),
        sign='>0',
        default=focal/3.,
    ))

    if collimator is False:
        collimator_length = None

    # -------------
    # diverging
    # -------------

    diverging = ds._generic_check._check_var(
        diverging, 'diverging',
        types=bool,
        default=False,
    )

    # -------------
    # kcase
    # -------------

    if collimator is True:
        if diverging is True:
            kcase = 'collimator_diverging'
        else:
            kcase = 'collimator_converging'
    else:
        kcase = 'pinhole'

    return (
        det_size,
        ap_size,
        focal,
        pitch,
        det_nb,
        dist,
        # equivalent collimator
        collimator,
        collimator_length,
        diverging,
        kcase,
    )


# ################################################
# ################################################
#           Compute detector
# ################################################


def _compute_sensors(
    det_nb=None,
    det_size=None,
    pitch=None,
    # unused
    **kwdargs,
):

    # ---------------
    # sensors
    # ---------------

    cy = pitch * np.linspace(-1, 1, det_nb)

    y_up = cy + det_size * 0.5
    y_low = cy - det_size * 0.5

    x_plot = np.zeros((det_nb*3))
    y_plot = np.array([y_low, y_up, np.full(cy.shape, np.nan)]).T.ravel()

    return {
        'cy': cy,
        'y_low': y_low,
        'y_up': y_up,
        'x_plot': x_plot,
        'y_plot': y_plot,
    }


# ################################################
# ################################################
#           Compute pinhole
# ################################################


def _compute_pinhole(
    dinputs=None,
    dsensors=None,
):

    # ---------------
    # extract
    # ---------------

    # inputs
    dist = dinputs['dist']
    focal = dinputs['focal']
    pitch = dinputs['pitch']
    det_size = dinputs['det_size']
    ap_size = dinputs['ap_size']

    # sensors
    cy = dsensors['cy']
    y_low = dsensors['y_low']
    y_up = dsensors['y_up']

    # ---------------
    # prepare & outputs
    # ---------------

    alpha = np.arctan(pitch / focal)
    beta = 2. * np.arctan((det_size + ap_size) / (2. * focal))
    R = (beta - alpha) / alpha

    # ---------------
    # aperture
    # ---------------

    apx = focal
    apy_up = 0.5 * ap_size
    apy_low = -0.5 * ap_size
    apx_plot = np.full((5,), apx)
    apy_plot = np.array([y_low[0], apy_low, np.nan, apy_up, y_up[-1]])

    # ---------------
    # LOS
    # ---------------

    vectx = apx
    vecty = -cy
    vectn = np.sqrt(vectx**2 + vecty**2)
    vectx = vectx / vectn
    vecty = vecty / vectn

    kplot = dist / vectx

    los_x = np.r_[0, dist]
    los_y = (
        cy[:, None]
        + kplot[:, None] * vecty[:, None] * np.r_[0, 1][None, :]
    )

    # ---------------
    # FOV
    # ---------------

    (
        fov_x,
        fov_in_y_low, fov_in_y_up,
        fov_out_y_low, fov_out_y_up,
    ) = _fov_inner_outer(
        dist=dist,
        # det
        y_low=y_low,
        y_up=y_up,
        # ap
        apx=apx,
        apy_low=apy_low,
        apy_up=apy_up,
    )

    # ---------------
    # dout
    # ---------------

    dout = {
        'outputs': {
            'alpha': alpha,
            'beta': beta,
            'R': R,
        },
        'aperture': {
            'cx': apx,
            'y_low': apy_low,
            'y_up': apy_up,
            'x_plot': apx_plot,
            'y_plot': apy_plot,
        },
        'LOS': {
            'vectx': vectx,
            'vecty': vecty,
            'x': los_x,
            'y': los_y,
        },
        'FOV': {
            'x': fov_x,
            'inner': {
                'y_low': fov_in_y_low,
                'y_up': fov_in_y_up,
            },
            'outer': {
                'y_low': fov_out_y_low,
                'y_up': fov_out_y_up,
            },
        },
    }

    return dout


# ################################################
# ################################################
#           Compute collimator - converging
# ################################################


def _compute_collimator_converging(
    dinputs=None,
    dsensors=None,
    dpinhole=None,
    # equivalent collimator
    length=None,
    diverging=None,
):

    # ---------------
    # extract
    # ---------------

    # inputs
    det_size = dinputs['det_size']
    det_nb = dinputs['det_nb']
    dist = dinputs['dist']

    # sensors
    det_y = dsensors['cy']
    dety_low = dsensors['y_low']
    dety_up = dsensors['y_up']

    # pinhole
    alpha = dpinhole['outputs']['alpha']
    R = dpinhole['outputs']['R']
    LOS_x = dpinhole['LOS']['vectx']
    LOS_y = dpinhole['LOS']['vecty']

    # ---------------
    # aperture
    # ---------------

    # ap_size to preserve overlap
    # beta = 2arctan((WA + WD) / (2 * length))
    # R = (beta - alpha) / alpha
    #
    # we want:
    # beta = alpha * (1 + R)
    # so:

    beta = alpha * (R + 1)
    ap_size = 2.*length * np.tan(beta / 2) - det_size

    # dist along LOS
    kk = length / LOS_x

    apx = length
    apy = det_y + LOS_y * kk

    apy_up = apy + 0.5 * ap_size
    apy_low = apy - 0.5 * ap_size

    ap_shape = (2*(det_nb+1) + det_nb,)
    apx_plot = np.full(ap_shape, apx)
    apy_plot = np.full(ap_shape, np.nan)
    apy_plot[0] = min(apy_low[0] - ap_size, det_y[0] - det_size*0.5)
    apy_plot[-1] = max(apy_up[-1] + ap_size, det_y[-1] + det_size*0.5)
    for ii in range(det_nb):
        i0 = 1 + ii*3
        i1 = 1 + ii*3 + 1
        i2 = 1 + ii*3 + 2
        apy_plot[i0] = apy[ii] - 0.5*ap_size
        apy_plot[i1] = np.nan
        apy_plot[i2] = apy[ii] + 0.5*ap_size

    # ----------------
    # LOS
    # ----------------

    # vectors
    vectx = apx
    vecty = apy - det_y
    vectn = np.sqrt(vectx**2 + vecty**2)
    vectx = vectx / vectn
    vecty = vecty / vectn

    # los
    kplot = dist / vectx
    los_x = np.r_[0, dist]
    los_y = (
        det_y[:, None]
        + kplot[:, None] * vecty[:, None] * np.r_[0, 1][None, :]
    )

    # ---------------
    # FOV
    # ---------------

    (
        fov_x,
        fov_in_y_low, fov_in_y_up,
        fov_out_y_low, fov_out_y_up,
    ) = _fov_inner_outer(
        dist=dist,
        # det
        y_low=dety_low,
        y_up=dety_up,
        # ap
        apx=apx,
        apy_low=apy_low,
        apy_up=apy_up,
    )

    # ---------------
    # dout
    # ---------------

    dout = {
        'outputs': {
            'ap_size': ap_size,
        },
        'aperture': {
            'cx': apx,
            'y_low': apy_low,
            'y_up': apy_up,
            'x_plot': apx_plot,
            'y_plot': apy_plot,
        },
        'LOS': {
            'vectx': vectx,
            'vecty': vecty,
            'x': los_x,
            'y': los_y,
        },
        'FOV': {
            'x': fov_x,
            'inner': {
                'y_low': fov_in_y_low,
                'y_up': fov_in_y_up,
            },
            'outer': {
                'y_low': fov_out_y_low,
                'y_up': fov_out_y_up,
            },
        },
    }

    return dout


# ################################################
# ################################################
#           Compute collimator - diverging
# ################################################


def _compute_collimator_diverging(
    dinputs=None,
    dsensors=None,
    dpinhole=None,
    # equivalent collimator
    length=None,
    diverging=None,
):

    # ---------------
    # extract
    # ---------------

    # inputs
    det_size = dinputs['det_size']
    det_nb = dinputs['det_nb']
    dist = dinputs['dist']

    # sensors
    det_y = dsensors['cy']
    dety_low = dsensors['y_low']
    dety_up = dsensors['y_up']

    # pinhole
    alpha = dpinhole['outputs']['alpha']
    R = dpinhole['outputs']['R']

    # ---------------
    # New LOS
    # ---------------

    los_x = dpinhole['LOS']['x']
    los_y = dpinhole['LOS']['y']
    los_y[:, -1] = los_y[::-1, -1]

    vectx = np.diff(los_x).ravel()
    vecty = np.diff(los_y, axis=1).ravel()
    vectn = np.sqrt(vectx**2 + vecty**2)
    vectx = vectx / vectn
    vecty = vecty / vectn

    # ---------------
    # aperture positions
    # ---------------

    apx = length
    kk = length / vectx
    apy = det_y + kk * vecty

    # ---------------
    # aperture
    # ---------------

    # ap_size to preserve overlap
    # beta = 2arctan((WA + WD) / (2 * length))
    # R = (beta - alpha) / alpha
    #
    # we want:
    # beta = alpha * (1 + R)
    # so:

    beta = alpha * (R + 1)
    ap_size = 2.*length * np.tan(beta / 2) - det_size

    apy_up = apy + 0.5 * ap_size
    apy_low = apy - 0.5 * ap_size

    ap_shape = (2*(det_nb+1) + det_nb,)
    apx_plot = np.full(ap_shape, apx)
    apy_plot = np.full(ap_shape, np.nan)
    apy_plot[0] = min(apy_low[0] - ap_size, det_y[0] - det_size*0.5)
    apy_plot[-1] = max(apy_up[-1] + ap_size, det_y[-1] + det_size*0.5)
    for ii in range(det_nb):
        i0 = 1 + ii*3
        i1 = 1 + ii*3 + 1
        i2 = 1 + ii*3 + 2
        apy_plot[i0] = apy[ii] - 0.5*ap_size
        apy_plot[i1] = np.nan
        apy_plot[i2] = apy[ii] + 0.5*ap_size

    # ---------------
    # FOV
    # ---------------

    (
        fov_x,
        fov_in_y_low, fov_in_y_up,
        fov_out_y_low, fov_out_y_up,
    ) = _fov_inner_outer(
        dist=dist,
        # det
        y_low=dety_low,
        y_up=dety_up,
        # ap
        apx=apx,
        apy_low=apy_low,
        apy_up=apy_up,
    )

    # ---------------
    # dout
    # ---------------

    dout = {
        'outputs': {
            'ap_size': ap_size,
        },
        'aperture': {
            'cx': apx,
            'y_low': apy_low,
            'y_up': apy_up,
            'x_plot': apx_plot,
            'y_plot': apy_plot,
        },
        'LOS': {
            'vectx': vectx,
            'vecty': vecty,
            'x': los_x,
            'y': los_y,
        },
        'FOV': {
            'x': fov_x,
            'inner': {
                'y_low': fov_in_y_low,
                'y_up': fov_in_y_up,
            },
            'outer': {
                'y_low': fov_out_y_low,
                'y_up': fov_out_y_up,
            },
        },
    }

    return dout


# ################################################
# ################################################
#           FOV - inner
# ################################################


def _fov_inner_outer(
    dist=None,
    # det
    y_low=None,
    y_up=None,
    # ap
    apx=None,
    apy_low=None,
    apy_up=None,
):

    fov_x = np.r_[0, dist]

    # ------
    # inner

    # vect low
    vlow_x = apx
    vlow_y = apy_low - y_low
    vlown = np.sqrt(vlow_x**2 + vlow_y**2)
    vlow_x, vlow_y = vlow_x/vlown, vlow_y/vlown

    # vect up
    vup_x = apx
    vup_y = apy_up - y_up
    vupn = np.sqrt(vup_x**2 + vup_y**2)
    vup_x, vup_y = vup_x/vupn, vup_y/vupn

    # y
    klow = dist / vlow_x
    kup = dist / vup_x
    fov_in_y_low = (
        y_low[:, None]
        + klow[:, None] * vlow_y[:, None] * np.r_[0, 1][None, :]
    )
    fov_in_y_up = (
        y_up[:, None]
        + kup[:, None] * vup_y[:, None] * np.r_[0, 1][None, :]
    )

    # ------
    # outer

    # vect low
    vlow_x = apx
    vlow_y = apy_up - y_low
    vlown = np.sqrt(vlow_x**2 + vlow_y**2)
    vlow_x, vlow_y = vlow_x/vlown, vlow_y/vlown

    # vect up
    vup_x = apx
    vup_y = apy_low - y_up
    vupn = np.sqrt(vup_x**2 + vup_y**2)
    vup_x, vup_y = vup_x/vupn, vup_y/vupn

    # y
    klow = dist / vlow_x
    kup = dist / vup_x
    fov_out_y_low = (
        y_low[:, None]
        + klow[:, None] * vlow_y[:, None] * np.r_[0, 1][None, :]
    )
    fov_out_y_up = (
        y_up[:, None]
        + kup[:, None] * vup_y[:, None] * np.r_[0, 1][None, :]
    )

    return fov_x, fov_in_y_low, fov_in_y_up, fov_out_y_low, fov_out_y_up


# ################################################
# ################################################
#           Add text - Inputs
# ################################################


def _add_text_inputs(dout, collimator=None):

    # -------------
    # inputs
    # -------------

    P = dout['inputs']['pitch']
    F = dout['inputs']['focal']
    WD = dout['inputs']['det_size']
    WA = dout['inputs']['ap_size']

    inputs = (
        r"$P = $" + f"{round(P*1000, ndigits=1)} mm\n"
        + r"$F = $" + f"{round(F*1000, ndigits=1)} mm\n"
        + r"$W_D = $" + f"{round(WD*1000, ndigits=1)} mm\n"
        + r"$W_A = $" + f"{round(WA*1000, ndigits=1)} mm\n"
    )

    if collimator is True:
        L = dout['inputs']['length']
        inputs += "\n" + r"$L = $" + f"{round(L*1000, ndigits=1)} mm\n"

    # -------------
    # store
    # -------------

    return {
        'tit_inputs': {
            'str': "Inputs",
            'pos': (0.10, 0.45),
            'fontweight': 'bold',
            'horizontalalignment': 'center',
        },
        'inputs': {
            'str': inputs,
            'pos': (0.05, 0.35),
        },
    }


# ################################################
# ################################################
#           Add text - Pinhole
# ################################################


def _add_text_pinhole(dout):

    # -------------
    # alpha, beta
    # -------------

    kpin = 'pinhole'
    alpha_str = round(dout[kpin]['outputs']['alpha']*180/np.pi, ndigits=1)
    alpha = (
        r"$\alpha = \arctan\left(\frac{P}{F}\right)$"
        + f" = {alpha_str} deg"
    )

    beta_str = round(dout[kpin]['outputs']['beta']*180/np.pi, ndigits=1)
    beta = (
        r"$\beta = 2\arctan\left(\frac{W_A + W_D}{2F}\right)$"
        + f" = {beta_str} deg"
    )

    # -------------
    # overlap rate
    # -------------

    R_str = round(dout[kpin]['outputs']['R']*100, ndigits=1)
    overlap = (
        "Overlap rate:\n    "
        + r"$R = \frac{\beta - \alpha}{\alpha}$"
        + r"$\approx \frac{W_A + W_D - P}{P}$"
        + "  " + r"$\approx$" + f"{R_str} %"
        + "\nOr:    "
        + r"$W_A \approx P(1 + R) - W_D$"
    )

    # -------------
    # Recommendation
    # -------------

    P = dout['inputs']['pitch']
    WD = dout['inputs']['det_size']
    WA_str = round(1.1*P - WD, ndigits=1)

    R10 = (
        "To get R = 10 %:\n    "
        + r"$W_A \approx 1.1P - W_D = $" + f"{WA_str} mm"
    )

    # -------------
    # Etendue
    # -------------

    etendue = (
        r"$E \approx \frac{(W_A.L_A)(W_D.L_D)}{F^2}$"
    )

    # -------------
    # store
    # -------------

    return {
        'tit_pinhole': {
            'str': 'PINHOLE',
            'pos': (0.35, 0.45),
            'fontweight': 'bold',
            'horizontalalignment': 'center',
        },
        'alpha': {
            'str': alpha,
            'pos': (0.25, 0.38),
        },
        'beta': {
            'str': beta,
            'pos': (0.25, 0.33),
        },
        'overlap': {
            'str': overlap,
            'pos': (0.25, 0.28),
        },
        'R10': {
            'str': R10,
            'pos': (0.25, 0.15),
        },
        'etendue': {
            'str': etendue,
            'pos': (0.25, 0.08),
        },
    }


# ################################################
# ################################################
#           Add text - Collimator - diverging
# ################################################


def _add_text_collimator_diverging(dout, kcase):

    # -------------
    # Aperture size
    # -------------

    WD = dout['inputs']['det_size']
    WA = dout['inputs']['ap_size']

    WA_str = round(dout[kcase]['outputs']['ap_size']*1000, ndigits=2)
    lim = round(WD / (WD+WA), ndigits=2)
    WA_str = (
        r"$W_A^C = 2.L\tan(\frac{\beta}{2}) - W_D$"
        r"$= \frac{L}{F}(W_A + W_D) - W_D$"
        + f"= {WA_str} mm"
        + "\n  Only if "
        + r"$L \geq F\frac{W_D}{W_A + W_D} \approx$" + f"{lim} F"
    )

    # -------------
    # etendue
    # -------------

    F = dout['inputs']['focal']
    L = dout['inputs']['length']
    WAC = dout[kcase]['outputs']['ap_size']

    coef = round((F/L)**2 * WAC/WA, ndigits=2)
    etendue = (
        r"$E^C \approx \frac{(W_A^C.L_A^C)(W_D.L_D)}{L^2}$"
        + r"$= \frac{F^2}{L^2} \frac{W_A^C.L_A^C}{W_A.L_A} E$"
        + r"$\approx$" + f"{coef}" + r"$\frac{L_A^C}{L_A}E$"
    )

    # -------------
    # store
    # -------------

    return {
        # titles
        'tit_collim': {
            'str': "COLLIMATOR EQUIVALENT\n(DIVERGING)",
            'pos': (0.80, 0.45),
            'fontweight': 'bold',
            'horizontalalignment': 'center',
        },
        # collimator
        'WA': {
            'str': WA_str,
            'pos': (0.60, 0.35),
        },
        # etendue
        'etendueC': {
            'str': etendue,
            'pos': (0.60, 0.20),
        },
    }


# ################################################
# ################################################
#           Add text - Collimator - converging
# ################################################


def _add_text_collimator_converging(dout, kcase):

    # -------------
    # Aperture size
    # -------------

    WD = dout['inputs']['det_size']
    WA = dout['inputs']['ap_size']

    WA_str = round(dout[kcase]['outputs']['ap_size']*1000, ndigits=2)
    lim = round(WD / (WD+WA), ndigits=2)
    WA_str = (
        r"$W_A^C = 2.L\tan(\frac{\beta}{2}) - W_D$"
        r"$= \frac{L}{F}(W_A + W_D) - W_D$"
        + f"= {WA_str} mm"
        + "\n  Only if "
        + r"$L \geq F\frac{W_D}{W_A + W_D} \approx$" + f"{lim} F"
    )

    # -------------
    # etendue
    # -------------

    F = dout['inputs']['focal']
    L = dout['inputs']['length']
    WAC = dout[kcase]['outputs']['ap_size']

    coef = round((F/L)**2 * WAC/WA, ndigits=2)
    etendue = (
        r"$E^C \approx \frac{(W_A^C.L_A^C)(W_D.L_D)}{L^2}$"
        + r"$= \frac{F^2}{L^2} \frac{W_A^C.L_A^C}{W_A.L_A} E$"
        + r"$\approx$" + f"{coef}" + r"$\frac{L_A^C}{L_A}E$"
    )

    # -------------
    # store
    # -------------

    return {
        # titles
        'tit_collim': {
            'str': "COLLIMATOR EQUIVALENT\n(CONVERGING)",
            'pos': (0.80, 0.45),
            'fontweight': 'bold',
            'horizontalalignment': 'center',
        },
        # collimator
        'WA': {
            'str': WA_str,
            'pos': (0.60, 0.35),
        },
        # etendue
        'etendueC': {
            'str': etendue,
            'pos': (0.60, 0.20),
        },
    }


# ################################################
# ################################################
#           Plot
# ################################################


def _plot(
    dout=None,
    kcase=None,
):

    # ----------------
    # prepare data
    # ----------------

    # for comparison
    kcase0 = 'pinhole'

    # ----------------
    # prepare figure
    # ----------------

    # dmargin
    dmargin = {
        'bottom': 0.1, 'top': 0.9,
        'left': 0.1, 'right': 0.9,
        'wspace': 0.1, 'hspace': 0.1,
    }

    # figure
    fig = plt.figure(figsize=(13, 9))
    gs = gridspec.GridSpec(ncols=1, nrows=2, **dmargin)

    # axes
    ax = fig.add_subplot(gs[0, :], aspect='equal', adjustable='datalim')
    ax.set_xlabel('X (m)', size=12, fontweight='bold')
    ax.set_ylabel('Y (m)', size=12, fontweight='bold')

    # ------------------------
    # plot sensors & aperture
    # ------------------------

    # sensors
    ax.plot(
        dout['sensors']['x_plot'],
        dout['sensors']['y_plot'],
        c='k',
        lw=3,
    )

    # aperture - pinhole
    ax.plot(
        dout[kcase0]['aperture']['x_plot'],
        dout[kcase0]['aperture']['y_plot'],
        c='k' if kcase == kcase0 else (0.8, 0.8, 0.8, 0.8),
        lw=3,
    )

    # apertures
    ax.plot(
        dout[kcase]['aperture']['x_plot'],
        dout[kcase]['aperture']['y_plot'],
        c='k',
        lw=3,
    )

    # ------------------------
    # plot LOS and FOV
    # ------------------------

    for ii in range(dout[kcase0]['LOS']['y'].shape[0]):

        # LOS - pinhole
        l0, = ax.plot(
            dout[kcase0]['LOS']['x'],
            dout[kcase0]['LOS']['y'][ii],
            ls='-' if kcase == kcase0 else '--',
            lw=1,
        )

        # LOS - case
        if dout[kcase].get('LOS') is not None:
            l0, = ax.plot(
                dout[kcase]['LOS']['x'],
                dout[kcase]['LOS']['y'][ii],
                ls='-',
                lw=1,
                c=l0.get_color(),
            )

        # FOV inner
        ax.fill_between(
            dout[kcase]['FOV']['x'],
            dout[kcase]['FOV']['inner']['y_low'][ii],
            dout[kcase]['FOV']['inner']['y_up'][ii],
            fc=l0.get_color(),
            alpha=0.6,
        )

        # FOV outer
        ax.fill_between(
            dout[kcase]['FOV']['x'],
            dout[kcase]['FOV']['outer']['y_low'][ii],
            dout[kcase]['FOV']['outer']['y_up'][ii],
            fc=l0.get_color(),
            alpha=0.2,
        )

    # ----------------
    # text
    # ----------------

    if dout.get('text') is not None:
        for k0, v0 in dout['text'].items():

            ax.text(
                v0['pos'][0],
                v0['pos'][1],
                v0['str'],
                size=14,
                fontweight=v0.get('fontweight', 'normal'),
                verticalalignment=v0.get('verticalalignment', 'top'),
                horizontalalignment=v0.get('horizontalalignment', 'left'),
                transform=fig.transFigure,
            )

        # figure vline
        x0 = 0.5*(
            dout['text']['tit_inputs']['pos'][0]
            + dout['text']['tit_pinhole']['pos'][0]
        )
        fig.add_artist(mlines.Line2D(
            [x0, x0],
            [0.05, 0.45],
            linewidth=2,
            color='k',
        ))

        # figure vline
        if kcase != kcase0:
            x0 = 0.5*(
                dout['text']['tit_pinhole']['pos'][0]
                + dout['text']['tit_collim']['pos'][0]
            )
            fig.add_artist(mlines.Line2D(
                [x0, x0],
                [0.05, 0.45],
                linewidth=2,
                color='k',
            ))

    return



# -*- coding: utf-8 -*-
"""
Created on Thu Sep 12 10:17:01 2024

@author: dvezinet
"""


import numpy as np
from matplotlib import path as mpath
import datastock as ds


from ..geom import _GG
from ._class00_poly2d_sample import main as poly2d_sample


# ###############################################################
# ###############################################################
#                      Main
# ###############################################################


def main(
    coll=None,
    key=None,
    # to sample on a single optics
    key_optics=None,
    # sampling
    dsampling_pixel=None,
    dsampling_optics=None,
    # optics (to restrain to certain optics only for faster)
    optics=None,
    # computing
    config=None,
    # storing
    store=None,
    strict=None,
    key_rays=None,
    overwrite=None,
):
    """

    Parameters
    ----------
    coll : TYPE, optional
        DESCRIPTION. The default is None.
    key : TYPE, optional
        DESCRIPTION. The default is None.
    # to sample on a single optics    key_optics : TYPE, optional
        DESCRIPTION. The default is None.
    # sampling    dsampling_pixel : TYPE, optional
        DESCRIPTION. The default is None.
    dsampling_optics : TYPE, optional
        DESCRIPTION. The default is None.
    # computing    config : TYPE, optional
        DESCRIPTION. The default is None.
    # storing    store : TYPE, optional
        DESCRIPTION. The default is None.
    overwrite : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    dout : TYPE
        DESCRIPTION.

    """

    # -------------
    # check
    # -------------

    (
     key,
     dsampling_pixel,
     dsampling_optics,
     optics,
     store, strict, key_rays, overwrite,
    ) = _check(
        coll=coll,
        key=key,
        # sampling
        dsampling_pixel=dsampling_pixel,
        dsampling_optics=dsampling_optics,
        # optics
        optics=optics,
        # storing
        store=store,
        strict=strict,
        key_rays=key_rays,
        overwrite=overwrite,
    )

    # ---------------
    #  prepare
    # ---------------

    wdiag = 'diagnostic'
    key_cam = coll.dobj[wdiag][key]['camera']
    doptics = coll.dobj[wdiag][key]['doptics']

    # ---------------
    #  compute
    # ---------------

    # initialize
    dout = {}

    # loop on cameras
    for kcam in key_cam:

        # --------------------
        # sample generic pixel

        dout[kcam] = _generic(
            coll=coll,
            key=key,
            kcam=kcam,
            doptics=doptics[kcam],
            # sampling
            dsampling_pixel=dsampling_pixel,
            dsampling_optics=dsampling_optics,
            # optics
            optics=optics.get(kcam),
        )

    # ---------------
    #  store
    # ---------------

    if store is True:
        _store(
            coll=coll,
            dout=dout,
            key_rays=key_rays,
            config=config,
            strict=strict,
            overwrite=overwrite,
        )

    else:
        return dout


# ###############################################################
# ###############################################################
#                      check
# ###############################################################


def _check(
    coll=None,
    key=None,
    # sampling
    dsampling_pixel=None,
    dsampling_optics=None,
    # optics
    optics=None,
    # storing
    store=None,
    strict=None,
    key_rays=None,
    overwrite=None,
):

    # ------------
    # key
    # ------------

    wdiag = 'diagnostic'
    lok = list(coll.dobj.get(wdiag, {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    key_cam = coll.dobj[wdiag][key]['camera']

    # spectro ?
    spectro = coll.dobj[wdiag][key]['spectro']
    if spectro is True:
        raise NotImplementedError()

    # ------------
    # dsampling_pixel
    # ------------

    ldict = [
        ('dsampling_pixel', dsampling_pixel),
        ('dsampling_optics', dsampling_optics),
    ]
    lk = [('dedge', ['res', 'factor']), ('dsurface', ['res', 'nb'])]
    kfunc = 'tf.data.poly2d_sample()'
    for ii, (kdict, vdict) in enumerate(ldict):
        c0 = (
            isinstance(vdict, dict)
            and any([kk in vdict.keys() for (kk, largs) in lk])
        )
        if not c0:
            lstr = [f"\t- {kk}: dict of {largs}" for (kk, largs) in lk]
            msg = (
                f"Arg '{kdict}' must be a dict with at least one of:\n"
                + "\n".join(lstr)
                + f"\nFed to {kfunc}\n"
                + f"\nProvided:\n{vdict}\n"
            )
            raise Exception(msg)

    # ------------
    # optics
    # ------------

    if optics is not None:

        # index => position in optics list
        optics0 = {
            kc: _check_optics_for_kcam(
                doptics=coll.dobj[wdiag][key]['doptics'][kc],
                optics=optics.get(kc) if isinstance(optics, dict) else optics,
                shape_cam=coll.dobj['camera'][kc]['dgeom']['shape'],
            )
            for kc in key_cam
        }

    else:
        optics0 = {}

    # -----------
    # store
    # ------------

    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=True,
    )

    # -----------
    # strict
    # ------------

    strict = ds._generic_check._check_var(
        strict, 'stict',
        types=bool,
        default=False,
    )

    # ------------
    # overwrite
    # ------------

    overwrite = ds._generic_check._check_var(
        overwrite, 'overwrite',
        types=bool,
        default=False,
    )

    # ------------
    # key_rays
    # ------------

    if store is True:

        if key_rays is None:
            key_rays = {kcam: f"{kcam}_rays" for kcam in key_cam}

        else:
            if len(key_cam) == 1 and isinstance(key_rays, str):
                key_rays = {key_cam[0]: key_rays}

        ncam = len(key_cam)
        c0 = (
            isinstance(key_rays, dict)
            and all([
                isinstance(key_rays.get(kcam), str)
                and len(set([key_rays[kcam] for kcam in key_cam])) == ncam
                and all([key_rays[kcam] not in coll.dobj['camera'].keys()])
                for kcam in key_cam
            ])
        )
        if not c0:
            msg = (
                "Arg key_rays must be a dict of key_ray for each camera:\n"
                f"\t- key_diag = '{key}'\n"
                f"\t- key_cam = {key_cam}\n"
                f"\t- key_rays = {key_rays}\n"
            )
            raise Exception(msg)

    else:
        key_rays = None

    return (
        key,
        dsampling_pixel, dsampling_optics,
        optics0,
        store, strict, key_rays, overwrite,
    )


def _check_optics_for_kcam(
    doptics=None,
    optics=None,
    shape_cam=None,
):

    if isinstance(optics, int):

        ind0 = np.arange(0, len(doptics['optics']))
        if doptics['pinhole'] is True:
            lok = np.r_[ind0, -1]
        else:
            nop = doptics['paths'].sum(axis=1)
            lok = np.r_[
                np.arange(0, np.min(nop)),
                -np.arange(1, np.min(nop)+1),
            ]

        optics = ds._generic_check._check_var(
            optics, 'optics',
            types=int,
            allowed=lok,
        )

        if doptics['pinhole'] is True:
            optics = [doptics['optics'][optics]]
        else:
            assert doptics.get('paths') is not None
            paths = doptics['paths']

            optics = [
                doptics['optics'][ind0[paths[ii, :]][optics]]
                for ii in range(shape_cam[0])
            ]

    # -------------------
    # key or list of keys

    if isinstance(optics, str):
        optics = [optics]

    optics = ds._generic_check._check_var_iter(
        optics, 'optics',
        types=(list, tuple),
        types_iter=str,
        allowed=doptics['optics'],
    )

    # eliminate redundants (ex. collimator with 1 common optic)
    return list(set(optics))


# ###############################################################
# ###############################################################
#                      random
# ###############################################################


def _random(
    coll=None,
    kcam=None,
    doptics=None,
    out_arr=None,
    nrays=None,
    strategy=None,
):

    # ---------------
    # prepare pixel
    # ---------------

    # outline path
    path = mpath.Path(out_arr)

    min0, max0 = np.min(out_arr[:, 0]), np.max(out_arr[:, 0])
    min1, max1 = np.min(out_arr[:, 1]), np.max(out_arr[:, 1])

    # areas to get safety factor
    area = coll.dobj['camera'][kcam]['dgeom']['pix_area']
    area_max = (max0 - min0) * (max1 - min1)

    # start point coords in pixel's plane
    start0, start1 = _seed_pix(
        path=path,
        nrays=nrays,
        nrays_factor=int(np.ceil(area_max/area) + 2),
        min0=min0,
        max0=max0,
        min1=min1,
        max1=max1,
    )

    # shared apertures ?
    pinhole = doptics['pinhole']
    parallel = coll.dobj['camera'][kcam]['dgeom']['parallel']

    # shape camera
    shape_cam = coll.dobj['camera'][kcam]['dgeom']['shape']

    # camera vector
    # get camera vectors
    dvect = coll.get_camera_unit_vectors(kcam)
    lc = ['x', 'y', 'z']
    if parallel is True:
        e0i_x, e0i_y, e0i_z = [dvect[f"e0_{kk}"] for kk in lc]
        e1i_x, e1i_y, e1i_z = [dvect[f"e1_{kk}"] for kk in lc]

    # -----------------------------------
    # prepare output
    # -----------------------------------

    shape_out = shape_cam + (nrays*nrays,)

    dout = {
        'key': f'{kcam}_rays_{strategy}',
        'start_x': np.full(shape_out, np.nan),
        'start_y': np.full(shape_out, np.nan),
        'start_z': np.full(shape_out, np.nan),
        'vect_x': np.full(shape_out, np.nan),
        'vect_y': np.full(shape_out, np.nan),
        'vect_z': np.full(shape_out, np.nan),
    }

    # -----------------------------------
    # pinhole camera (shared apertures)
    # -----------------------------------

    if pinhole is True:

        # get pixel centers
        cx, cy, cz = coll.get_camera_cents_xyz(kcam)
        cent_cam = np.r_[np.mean(cx), np.mean(cy), np.mean(cz)]

        # get end points
        end0, end1, op_cent, op_e0, op_e1 = _seed_optics(
            coll=coll,
            cent_cam=cent_cam,
            nrays=nrays,
            optics=doptics['optics'],
        )

        # full versions
        end0f = np.tile(end0, nrays)
        end1f = np.tile(end1, nrays)

        # full versions
        start0f = np.repeat(start0, end0.size)
        start1f = np.repeat(start1, end0.size)

        # vignetting polygons
        # vignett_poly = [
            # np.array(coll.get_optics_poly(k0, closed=True, add_points=False))
            # for k0 in doptics['optics']
        # ]

        # loop on pixels
        ray_orig = np.full((3, nrays*end0.size), np.nan)
        ray_vdir = np.full((3, nrays*end0.size), np.nan)
        for ind in np.ndindex(shape_cam):

            sli_out = ind + (slice(None),)

            # unit vectors
            if parallel is not True:
                e0i_x, e0i_y, e0i_z = [dvect[f"e0_{kk}"][ind] for kk in lc]
                e1i_x, e1i_y, e1i_z = [dvect[f"e1_{kk}"][ind] for kk in lc]

            # ray_orig
            ray_orig[0, :] = cx[ind] + start0f * e0i_x + start1f * e1i_x
            ray_orig[1, :] = cy[ind] + start0f * e0i_y + start1f * e1i_y
            ray_orig[2, :] = cz[ind] + start0f * e0i_z + start1f * e1i_z

            # end points
            ray_vdir[0, :] = op_cent[0] + end0f * op_e0[0] + end1f * op_e1[0]
            ray_vdir[1, :] = op_cent[1] + end0f * op_e0[1] + end1f * op_e1[1]
            ray_vdir[2, :] = op_cent[2] + end0f * op_e0[2] + end1f * op_e1[2]

            # ray_vdir, normalized
            ray_vdir[0, :] = ray_vdir[0, :] - ray_orig[0, :]
            ray_vdir[1, :] = ray_vdir[1, :] - ray_orig[1, :]
            ray_vdir[2, :] = ray_vdir[2, :] - ray_orig[2, :]

            ray_norm = np.sqrt(np.sum(ray_vdir**2, axis=0))
            ray_vdir[:] = ray_vdir / ray_norm

            # vignetting (npoly, nrays) is wrong !!!
            # iok = _GG.vignetting(
            #     ray_orig,
            #     ray_vdir,
            #     vignett_poly,
            #     lnvert,
            #     num_threads=16,
            # )
            # print(ind, iok.sum(), iok.size)

            indrand = (np.random.randint(0, ray_orig.shape[1], nrays*nrays),)

            dout['start_x'][sli_out] = ray_orig[(0,) + indrand]
            dout['start_y'][sli_out] = ray_orig[(1,) + indrand]
            dout['start_z'][sli_out] = ray_orig[(2,) + indrand]
            dout['vect_x'][sli_out] = ray_vdir[(0,) + indrand]
            dout['vect_y'][sli_out] = ray_vdir[(1,) + indrand]
            dout['vect_z'][sli_out] = ray_vdir[(2,) + indrand]

    # ---------------
    # mesh pixels
    # ---------------

    else:

        for ind in np.ndindex(shape_cam):

            print(ind)

    return dout


# #########################
# seeding pixels
# #########################


def _seed_pix(
    path=None,
    nrays=None,
    nrays_factor=None,
    min0=None,
    max0=None,
    min1=None,
    max1=None,
):

    seed0 = np.random.random((nrays*nrays_factor,))
    seed1 = np.random.random((nrays*nrays_factor,))

    pts0 = (max0 - min0) * seed0 + min0
    pts1 = (max1 - min1) * seed1 + min1

    # only keep those in the pixel
    iok_pix = path.contains_points(np.array([pts0, pts1]).T).nonzero()[0]

    # imax
    imax = min(nrays, iok_pix.size)

    return pts0[iok_pix[:imax]], pts1[iok_pix[:imax]]


# #########################
# seeding optics
# #########################


def _seed_optics(
    coll=None,
    cent_cam=None,
    nrays=None,
    optics=None,
):

    # -----------------------------
    # select smallest etendue optic
    # -----------------------------

    optics, optics_cls = coll.get_optics_cls(optics)

    etend = [
        (
            coll.dobj[cc][kop]['dgeom']['area']
            / np.linalg.norm(coll.dobj[cc][kop]['dgeom']['cent'] - cent_cam)**2
        )
        for kop, cc in zip(optics, optics_cls)
    ]

    ind = np.nanargmin(etend)
    kop = optics[ind]
    cc = optics_cls[ind]

    # -----------------------------
    # seed
    # -----------------------------

    out0, out1 = coll.get_optics_outline(kop, add_points=False, closed=False)

    path = mpath.Path(np.array([out0, out1]).T)

    min0, max0 = np.min(out0), np.max(out0)
    min1, max1 = np.min(out1), np.max(out1)

    area = coll.dobj[cc][kop]['dgeom']['area']
    factor = int(np.ceil((max0 - min0) * (max1 - min1) / area) + 2)

    seed0 = np.random.random((nrays * factor,))
    seed1 = np.random.random((nrays * factor,))

    pts0 = (max0 - min0) * seed0 + min0
    pts1 = (max1 - min1) * seed1 + min1

    # only keep those in the pixel
    iok_pix = path.contains_points(np.array([pts0, pts1]).T).nonzero()[0]

    # imax
    imax = min(nrays*2, iok_pix.size)

    # optic cent and vectors
    cent = coll.dobj[cc][kop]['dgeom']['cent']
    e0 = coll.dobj[cc][kop]['dgeom']['e0']
    e1 = coll.dobj[cc][kop]['dgeom']['e1']

    return pts0[iok_pix[:imax]], pts1[iok_pix[:imax]], cent, e0, e1


# ###############################################################
# ###############################################################
#                      generic
# ###############################################################


def _generic(
    coll=None,
    key=None,
    kcam=None,
    doptics=None,
    # sampling options
    dsampling_pixel=None,
    dsampling_optics=None,
    # optics
    optics=None,
):

    # -------------------
    # optics
    # -------------------

    if optics is None:
        optics = doptics['optics']

    # ---------------------
    # generic pixel outline
    # ---------------------

    # get pixel outline
    kout0, kout1 = coll.dobj['camera'][kcam]['dgeom']['outline']

    # sample pixel
    dout_pixel = poly2d_sample(
        coll.ddata[kout0]['data'],
        coll.ddata[kout1]['data'],
        key=kcam,
        dedge=dsampling_pixel.get('dedge'),
        dsurface=dsampling_pixel.get('dsurface'),
    )

    nstart = dout_pixel['x0'].size

    # ---------------------
    # prepare
    # ---------------------

    # shared apertures ?
    parallel = coll.dobj['camera'][kcam]['dgeom']['parallel']

    # shape camera
    shape_cam = coll.dobj['camera'][kcam]['dgeom']['shape']
    shape_start = shape_cam + (nstart,)

    # camera vector
    # get camera vectors
    dvect = coll.get_camera_unit_vectors(kcam)
    lc = ['x', 'y', 'z']
    if parallel is True:
        e0i_x, e0i_y, e0i_z = [dvect[f"e0_{kk}"] for kk in lc]
        e1i_x, e1i_y, e1i_z = [dvect[f"e1_{kk}"] for kk in lc]

    cents_x, cents_y, cents_z = coll.get_camera_cents_xyz(kcam)

    # -----------------------
    # prepare buffer for cx, cy, cz
    # -----------------------

    # cx, cy, cz (3d poitns for each pixel)
    cx = np.full((nstart,), np.nan)
    cy = np.full((nstart,), np.nan)
    cz = np.full((nstart,), np.nan)

    # ---------------------
    # Only 1 optic (pinhole or collimator with 1 common optics)
    # ---------------------

    if len(optics) == 1:

        # ------------------------
        # get end points on optics

        endx, endy, endz = _get_end_optics(
            coll=coll,
            kop=optics[0],
            dsampling_optics=dsampling_optics,
        )
        nend = endx.size

        # --------------
        # prepare output

        # rays shape for each pixel
        shape_each = (nstart, nend)
        vx = np.full(shape_each, np.nan)
        vy = np.full(shape_each, np.nan)
        vz = np.full(shape_each, np.nan)

        # oreall
        shape_out = shape_start + (nend,)

        dout = {
            'start_x': np.full(shape_out, np.nan),
            'start_y': np.full(shape_out, np.nan),
            'start_z': np.full(shape_out, np.nan),
            'vect_x': np.full(shape_out, np.nan),
            'vect_y': np.full(shape_out, np.nan),
            'vect_z': np.full(shape_out, np.nan),
            # 'dsang': np.full(shape_out, np.nan),
        }

        # -------------------
        # assign output

        for ii, ind in enumerate(np.ndindex(shape_cam)):

            if not parallel:
                e0i_x, e0i_y, e0i_z = [dvect[f"e0_{kk}"][ind] for kk in lc]
                e1i_x, e1i_y, e1i_z = [dvect[f"e1_{kk}"][ind] for kk in lc]

            # update cx, cy, cz
            _update_cxyz(
                parallel,
                e0i_x, e0i_y, e0i_z,
                e1i_x, e1i_y, e1i_z,
                cents_x, cents_y, cents_z,
                ind, lc,
                dout_pixel,
                cx, cy, cz,
                dvect,
            )

            # vect
            vx[...] = endx[None, :] - cx[:, None]
            vy[...] = endy[None, :] - cy[:, None]
            vz[...] = endz[None, :] - cz[:, None]

            vnorm_inv = 1./np.sqrt(vx**2 + vy**2 + vz**2)

            # slicing
            sli = ind + (slice(None), slice(None))

            # assigning
            dout['start_x'][sli] = np.copy(cx)[:, None]
            dout['start_y'][sli] = np.copy(cy)[:, None]
            dout['start_z'][sli] = np.copy(cz)[:, None]

            dout['vect_x'][sli] = vx * vnorm_inv
            dout['vect_y'][sli] = vy * vnorm_inv
            dout['vect_z'][sli] = vz * vnorm_inv

    # ---------------------
    # loop on pixels
    # ---------------------

    else:

        dout = {
            'start_x': {},
            'start_y': {},
            'start_z': {},
            'vect_x': {},
            'vect_y': {},
            'vect_z': {},
            # 'dsang': np.full(shape_out, np.nan),
        }

        for ii, ind in enumerate(np.ndindex(shape_cam)):

            # check optics per pixel
            lop = np.array(doptics['optics'])[doptics['paths'][ii, :]]
            lop = [
                kop for kop in lop
                if (kop in optics or optics is None)
            ]

            # check number of optics
            if len(lop) > 1:
                msg = (
                    "Multiple optics not handled yet!\n"
                    f"\t- key_diag = '{key}'\n"
                    f"\t- kcam = '{kcam}'\n"
                    f"\t- ii, ind = {ii}, {ind}\n"
                    f"\t- optics = {optics}\n"
                    f"\t- lop = {lop}\n"
                )
                raise NotImplementedError(msg)

            # ------------------------
            # get end points on optics

            endx, endy, endz = _get_end_optics(
                coll=coll,
                kop=lop[0],
                dsampling_optics=dsampling_optics,
            )
            nend = endx.size

            # ------------------
            # update pixel cx, cy, cz

            if not parallel:
                e0i_x, e0i_y, e0i_z = [dvect[f"e0_{kk}"][ind] for kk in lc]
                e1i_x, e1i_y, e1i_z = [dvect[f"e1_{kk}"][ind] for kk in lc]

            _update_cxyz(
                parallel,
                e0i_x, e0i_y, e0i_z,
                e1i_x, e1i_y, e1i_z,
                cents_x, cents_y, cents_z,
                ind, lc,
                dout_pixel,
                cx, cy, cz,
                dvect,
            )

            # vector
            vx = endx[None, :] - cx[:, None]
            vy = endy[None, :] - cy[:, None]
            vz = endz[None, :] - cz[:, None]

            vnorm_inv = 1./np.sqrt(vx**2 + vy**2 + vz**2)

            # assigning
            dout['start_x'][ind] = np.copy(cx)[:, None]
            dout['start_y'][ind] = np.copy(cy)[:, None]
            dout['start_z'][ind] = np.copy(cz)[:, None]

            dout['vect_x'][ind] = vx * vnorm_inv
            dout['vect_y'][ind] = vy * vnorm_inv
            dout['vect_z'][ind] = vz * vnorm_inv

    # -------------
    # adjust
    # -------------

    if isinstance(dout['start_x'], dict):

        dnrays = {k0: v0.shape[1] for k0, v0 in dout['vect_x'].items()}
        nraysu = np.unique([v0 for v0 in dnrays.values()])

        lkstart = [f'start_{cc}' for cc in lc]
        lkvect = [f'vect_{cc}' for cc in lc]

        # initialize
        for kk in lkstart + lkvect:
            oo = np.full(shape_cam + (nstart, nraysu.max()), np.nan)

            for ii, ind in enumerate(np.ndindex(shape_cam)):
                sli = ind + (slice(None), slice(0, dnrays[ind], 1))
                oo[sli] = dout[kk][ind]

            dout[kk] = oo

    # ---------------
    # return
    # ---------------

    return dout


def _get_end_optics(
    coll=None,
    kop=None,
    dsampling_optics=None,
):

    kop, clsop = coll.get_optics_cls(kop)
    clsop, kop = clsop[0], kop[0]
    kout0, kout1 = coll.dobj[clsop][kop]['dgeom']['outline']

    dout_optics = poly2d_sample(
        coll.ddata[kout0]['data'],
        coll.ddata[kout1]['data'],
        key=kop,
        dedge=dsampling_optics.get('dedge'),
        dsurface=dsampling_optics.get('dsurface'),
    )

    # get 3d optics end points coordinates
    func = coll.get_optics_x01toxyz(key=kop, asplane=False)
    return func(dout_optics['x0'], dout_optics['x1'])


def _update_cxyz(
    parallel,
    e0i_x, e0i_y, e0i_z,
    e1i_x, e1i_y, e1i_z,
    cents_x, cents_y, cents_z,
    ind, lc,
    dout_pixel,
    cx, cy, cz,
    dvect,
):

    if parallel is not True:
        e0i_x, e0i_y, e0i_z = [dvect[f"e0_{kk}"][ind] for kk in lc]
        e1i_x, e1i_y, e1i_z = [dvect[f"e1_{kk}"][ind] for kk in lc]

    # start
    cx[...] = (
        cents_x[ind]
        + dout_pixel['x0'] * e0i_x
        + dout_pixel['x1'] * e1i_x
    )
    cy[...] = (
        cents_y[ind]
        + dout_pixel['x0'] * e0i_y
        + dout_pixel['x1'] * e1i_y
    )
    cz[...] = (
        cents_z[ind]
        + dout_pixel['x0'] * e0i_z
        + dout_pixel['x1'] * e1i_z
    )

    return


# ###############################################################
# ###############################################################
#                      store
# ###############################################################


def _store(
    coll=None,
    kdiag=None,
    dout=None,
    key_rays=None,
    config=None,
    strict=None,
    overwrite=None,
):

    # --------------
    # store
    # --------------

    for kcam, v0 in dout.items():

        key = key_rays[kcam]

        if key in coll.dobj['rays'].keys() and overwrite is True:
            coll.remove_rays(key)

        # -----------------
        # add ref

        nstart, nends = v0['vect_x'].shape[-2:]

        krstart = f"{key}_nstart"
        coll.add_ref(key=krstart, size=nstart)

        krend = f"{key}_nend"
        coll.add_ref(key=krend, size=nends)

        # -----------------
        # add rays

        ref = coll.dobj['camera'][kcam]['dgeom']['ref'] + (krstart, krend)

        coll.add_rays(
            key=key,
            ref=ref,
            config=config,
            strict=strict,
            **v0
        )

    return



import sys
import os
import getpass
import argparse


# tofu
# test if in a tofu git repo
_HERE = os.path.abspath(os.path.dirname(__file__))
_TOFUPATH = os.path.dirname(_HERE)


def get_mods():
    istofugit = False
    if '.git' in os.listdir(_TOFUPATH) and 'tofu' in _TOFUPATH:
        istofugit = True

    if istofugit:
        # Make sure we load the corresponding tofu
        sys.path.insert(1, _TOFUPATH)
        import tofu as tf
        from tofu.imas2tofu import MultiIDSLoader
        _ = sys.path.pop(1)
    else:
        import tofu as tf
        from tofu.imas2tofu import MultiIDSLoader

    # default parameters
    pfe = os.path.join(os.path.expanduser('~'), '.tofu', '_scripts_def.py')
    if os.path.isfile(pfe):
        # Make sure we load the user-specific file
        # sys.path method
        # sys.path.insert(1, os.path.join(os.path.expanduser('~'), '.tofu'))
        # import _scripts_def as _defscripts
        # _ = sys.path.pop(1)
        # importlib method
        import importlib.util
        spec = importlib.util.spec_from_file_location("_defscripts", pfe)
        _defscripts = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(_defscripts)
    else:
        try:
            import tofu.entrypoints._def as _defscripts
        except Exception as err:
            from . import _def as _defscripts
    return tf, MultiIDSLoader, _defscripts


# #############################################################################
#       utility functions
# #############################################################################


def _str2bool(v):
    if isinstance(v, bool):
        return v
    elif v.lower() in ['yes', 'true', 'y', 't', '1']:
        return True
    elif v.lower() in ['no', 'false', 'n', 'f', '0']:
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected!')


def _str2boolstr(v):
    if isinstance(v, bool):
        return v
    elif isinstance(v, str):
        if v.lower() in ['yes', 'true', 'y', 't', '1']:
            return True
        elif v.lower() in ['no', 'false', 'n', 'f', '0']:
            return False
        elif v.lower() == 'none':
            return None
        else:
            return v
    else:
        raise argparse.ArgumentTypeError('Boolean, None or str expected!')


def _str2tlim(v):
    c0 = (v.isdigit()
          or ('.' in v
              and len(v.split('.')) == 2
              and all([vv.isdigit() for vv in v.split('.')])))
    if c0 is True:
        v = float(v)
    elif v.lower() == 'none':
        v = None
    return v


# #############################################################################
#       Parser for tofu version
# #############################################################################


def parser_version():
    msg = """ Get tofu version from bash optionally set an enviroment variable

    If run from a git repo containing tofu, simply returns git describe
    Otherwise reads the tofu version stored in tofu/version.py

    """
    ddef = {
        'path': os.path.join(_TOFUPATH, 'tofu'),
        'envvar': False,
        'verb': True,
        'warn': True,
        'force': False,
        'name': 'TOFU_VERSION',
    }

    # Instanciate parser
    parser = argparse.ArgumentParser(description=msg)

    # Define input arguments
    parser.add_argument('-p', '--path',
                        type=str,
                        help='tofu source directory where version.py is found',
                        required=False, default=ddef['path'])
    parser.add_argument('-v', '--verb',
                        type=_str2bool,
                        help='flag indicating whether to print the version',
                        required=False, default=ddef['verb'])
    parser.add_argument('-ev', '--envvar',
                        type=_str2boolstr,
                        help='name of the environment variable to set, if any',
                        required=False, default=ddef['envvar'])
    parser.add_argument('-w', '--warn',
                        type=_str2bool,
                        help=('flag indicatin whether to print a warning when'
                              + 'the desired environment variable (envvar)'
                              + 'already exists'),
                        required=False, default=ddef['warn'])
    parser.add_argument('-f', '--force',
                        type=_str2bool,
                        help=('flag indicating whether to force the update of '
                              + 'the desired environment variable (envvar)'
                              + ' even if it already exists'),
                        required=False, default=ddef['force'])

    return ddef, parser


# #############################################################################
#       Parser for tofu custom
# #############################################################################


def parser_custom():
    msg = """ Create a local copy of tofu default parameters

    This creates a local copy, in your home, of tofu default parameters
    A directory .tofu is created in your home directory
    In this directory, modules containing default parameters are copied
    You can then customize them without impacting other users

    """
    _USER = getpass.getuser()
    _USER_HOME = os.path.expanduser('~')

    ddef = {
        'target': os.path.join(_USER_HOME, '.tofu'),
        'source': os.path.join(_TOFUPATH, 'tofu'),
        'files': [
            '_imas2tofu_def.py',
            '_entrypoints_def.py',
        ],
        'directories': [
            'openadas2tofu',
            'nist2tofu',
            os.path.join('nist2tofu', 'ASD'),
        ],
    }

    # Instanciate parser
    parser = argparse.ArgumentParser(description=msg)

    # Define input arguments
    parser.add_argument('-s', '--source',
                        type=str,
                        help='tofu source directory',
                        required=False,
                        default=ddef['source'])
    parser.add_argument('-t', '--target',
                        type=str,
                        help=('directory where .tofu/ should be created'
                              + ' (default: {})'.format(ddef['target'])),
                        required=False,
                        default=ddef['target'])
    parser.add_argument('-f', '--files',
                        type=str,
                        help='list of files to be copied',
                        required=False,
                        nargs='+',
                        default=ddef['files'],
                        choices=ddef['files'])
    return ddef, parser


# #############################################################################
#       Parser for tofu plot
# #############################################################################


def parser_plot():

    tf, MultiIDSLoader, _defscripts = get_mods()

    _LIDS_CONFIG = MultiIDSLoader._lidsconfig
    _LIDS_DIAG = MultiIDSLoader._lidsdiag
    _LIDS_PLASMA = tf.imas2tofu.MultiIDSLoader._lidsplasma
    _LIDS = _LIDS_CONFIG + _LIDS_DIAG + _LIDS_PLASMA + tf.utils._LIDS_CUSTOM

    msg = """Fast interactive visualization tool for diagnostics data in
    imas

    This is merely a wrapper around the function tofu.load_from_imas()
    It loads (from imas) and displays diagnostics data from the following
    ids:
        {}
    """.format(_LIDS)

    ddef = {
        # User-customizable
        'run': _defscripts._TFPLOT_RUN,
        'user': _defscripts._TFPLOT_USER,
        'database': _defscripts._TFPLOT_DATABASE,
        'version': _defscripts._TFPLOT_VERSION,
        't0': _defscripts._TFPLOT_T0,
        'tlim': None,
        'sharex': _defscripts._TFPLOT_SHAREX,
        'bck': _defscripts._TFPLOT_BCK,
        'extra': _defscripts._TFPLOT_EXTRA,
        'indch_auto': _defscripts._TFPLOT_INDCH_AUTO,

        'config': _defscripts._TFPLOT_CONFIG,
        'tosoledge3x': _defscripts._TFPLOT_TOSOLEDGE3X,

        'mag_sep_nbpts': _defscripts._TFPLOT_MAG_SEP_NBPTS,
        'mag_sep_dR': _defscripts._TFPLOT_MAG_SEP_DR,
        'mag_init_pts': _defscripts._TFPLOT_MAG_INIT_PTS,

        # Non user-customizable
        'lids_plasma': _LIDS_PLASMA,
        'lids_diag': _LIDS_DIAG,
        'lids': _LIDS,
    }

    parser = argparse.ArgumentParser(description=msg)

    msg = 'config name to be loaded'
    parser.add_argument('-c', '--config', help=msg,
                        required=False, type=str,
                        default=ddef['config'])
    msg = 'path in which to save the tofu config in SOLEDGE3X format'
    parser.add_argument('-tse3x', '--tosoledge3x', help=msg,
                        required=False, type=str,
                        default=ddef['tosoledge3x'])
    parser.add_argument('-s', '--shot', type=int,
                        help='shot number', required=False, nargs='+')
    msg = 'username of the DB where the datafile is located'
    parser.add_argument('-u', '--user', help=msg, required=False,
                        default=ddef['user'])
    msg = 'database name where the datafile is located'
    parser.add_argument('-db', '--database', help=msg, required=False,
                        default=ddef['database'])
    parser.add_argument('-r', '--run', help='run number',
                        required=False, type=int,
                        default=ddef['run'])
    parser.add_argument('-v', '--version', help='version number',
                        required=False, type=str,
                        default=ddef['version'])

    msg = ("ids from which to load diagnostics data,"
           + " can be:\n{}".format(ddef['lids']))
    parser.add_argument('-i', '--ids', type=str, required=True,
                        help=msg, nargs='+', choices=ddef['lids'])
    parser.add_argument('-q', '--quantity', type=str, required=False,
                        help='Desired quantity from the plasma ids',
                        nargs='+', default=None)
    parser.add_argument('-X', '--X', type=str, required=False,
                        help='Quantity from the plasma ids for abscissa',
                        nargs='+', default=None)
    parser.add_argument('-t0', '--t0', type=_str2boolstr, required=False,
                        help='Reference time event setting t = 0',
                        default=ddef['t0'])
    parser.add_argument('-t', '--t', type=float, required=False,
                        help='Input time when needed')
    parser.add_argument('-tl', '--tlim', type=_str2tlim,
                        required=False,
                        help='limits of the time interval',
                        nargs='+', default=ddef['tlim'])
    parser.add_argument('-ich', '--indch', type=int, required=False,
                        help='indices of channels to be loaded',
                        nargs='+', default=None)
    parser.add_argument('-ichauto', '--indch_auto', type=_str2bool,
                        required=False,
                        help='automatically determine indices of'
                        + ' channels to be loaded', default=ddef['indch_auto'])
    parser.add_argument('-e', '--extra', type=_str2bool, required=False,
                        help='If True loads separatrix and heating power',
                        default=ddef['extra'])
    parser.add_argument('-sx', '--sharex', type=_str2bool, required=False,
                        help='Should X axis be shared between diag ids ?',
                        default=ddef['sharex'], const=True, nargs='?')
    parser.add_argument('-bck', '--background', type=_str2bool, required=False,
                        help='Plot data enveloppe as grey background ?',
                        default=ddef['bck'], const=True, nargs='?')

    parser.add_argument('-mag_sep_dR', '--mag_sep_dR', type=float,
                        required=False,
                        default=ddef['mag_sep_dR'],
                        help='Distance to separatrix from r_ext to plot'
                        + ' magnetic field lines')
    parser.add_argument('-mag_sep_nbpts', '--mag_sep_nbpts', type=int,
                        required=False,
                        default=ddef['mag_sep_nbpts'],
                        help=('Number of mag. field lines to plot '
                              + 'from separatrix'))
    parser.add_argument('-mag_init_pts', '--mag_init_pts',
                        type=float, required=False, nargs=3,
                        default=ddef['mag_init_pts'],
                        help='Manual coordinates of point that a RED magnetic'
                        + ' field line will cross on graphics,'
                        + ' give coordinates as: R [m], Phi [rad], Z [m]')
    return ddef, parser


# #############################################################################
#       Parser for tofu calc
# #############################################################################


def parser_calc():

    tf, MultiIDSLoader, _defscripts = get_mods()

    _LIDS_DIAG = MultiIDSLoader._lidsdiag
    _LIDS_PLASMA = tf.imas2tofu.MultiIDSLoader._lidsplasma
    _LIDS = _LIDS_DIAG + _LIDS_PLASMA + tf.utils._LIDS_CUSTOM

    # Parse input arguments
    msg = """Fast interactive visualization tool for diagnostics data in
    imas

    This is merely a wrapper around the function tofu.calc_from_imas()
    It calculates and dsplays synthetic signal (from imas) from the following
    ids:
        {}
    """.format(_LIDS)

    ddef = {
        # User-customizable
        'run': _defscripts._TFCALC_RUN,
        'user': _defscripts._TFCALC_USER,
        'database': _defscripts._TFCALC_DATABASE,
        'version': _defscripts._TFCALC_VERSION,
        't0': _defscripts._TFCALC_T0,
        'tlim': None,
        'sharex': _defscripts._TFCALC_SHAREX,
        'bck': _defscripts._TFCALC_BCK,
        'extra': _defscripts._TFCALC_EXTRA,
        'indch_auto': _defscripts._TFCALC_INDCH_AUTO,
        'coefs': None,

        # Non user-customizable
        'lids_plasma': _LIDS_PLASMA,
        'lids_diag': _LIDS_DIAG,
        'lids': _LIDS,
    }

    parser = argparse.ArgumentParser(description=msg)

    # Main idd parameters
    parser.add_argument('-s', '--shot', type=int,
                        help='shot number', required=True)
    msg = 'username of the DB where the datafile is located'
    parser.add_argument('-u', '--user',
                        help=msg, required=False, default=ddef['user'])
    msg = 'database name where the datafile is located'
    parser.add_argument('-db', '--database', help=msg, required=False,
                        default=ddef['database'])
    parser.add_argument('-r', '--run', help='run number',
                        required=False, type=int, default=ddef['run'])
    parser.add_argument('-v', '--version', help='version number',
                        required=False, type=str, default=ddef['version'])

    # Equilibrium idd parameters
    parser.add_argument('-s_eq', '--shot_eq', type=int,
                        help='shot number for equilibrium, defaults to -s',
                        required=False, default=None)
    msg = 'username for the equilibrium, defaults to -u'
    parser.add_argument('-u_eq', '--user_eq',
                        help=msg, required=False, default=None)
    msg = 'database name for the equilibrium, defaults to -tok'
    parser.add_argument('-db_eq', '--database_eq',
                        help=msg, required=False, default=None)
    parser.add_argument('-r_eq', '--run_eq',
                        help='run number for the equilibrium, defaults to -r',
                        required=False, type=int, default=None)

    # Profile idd parameters
    parser.add_argument('-s_prof', '--shot_prof', type=int,
                        help='shot number for profiles, defaults to -s',
                        required=False, default=None)
    msg = 'username for the profiles, defaults to -u'
    parser.add_argument('-u_prof', '--user_prof',
                        help=msg, required=False, default=None)
    msg = 'database name for the profiles, defaults to -tok'
    parser.add_argument('-db_prof', '--database_prof',
                        help=msg, required=False, default=None)
    parser.add_argument('-r_prof', '--run_prof',
                        help='run number for the profiles, defaults to -r',
                        required=False, type=int, default=None)

    msg = ("ids from which to load diagnostics data,"
           + " can be:\n{}".format(ddef['lids']))
    parser.add_argument('-i', '--ids', type=str, required=True,
                        help=msg, nargs='+', choices=ddef['lids'])
    parser.add_argument('-B', '--Brightness', type=bool, required=False,
                        help='Whether to express result as brightness',
                        default=None)
    parser.add_argument('-res', '--res', type=float, required=False,
                        help='Space resolution for the LOS-discretization',
                        default=None)
    parser.add_argument('-t0', '--t0', type=_str2boolstr, required=False,
                        help='Reference time event setting t = 0',
                        default=ddef['t0'])
    parser.add_argument('-tl', '--tlim', type=_str2tlim,
                        required=False,
                        help='limits of the time interval',
                        nargs='+', default=ddef['tlim'])
    parser.add_argument('-c', '--coefs', type=float, required=False,
                        help='Corrective coefficient, if any',
                        default=ddef['coefs'])
    parser.add_argument('-ich', '--indch', type=int, required=False,
                        help='indices of channels to be loaded',
                        nargs='+', default=None)
    parser.add_argument('-ichauto', '--indch_auto', type=bool, required=False,
                        help=('automatically determine indices '
                              + 'of channels to be loaded'),
                        default=ddef['indch_auto'])
    parser.add_argument('-e', '--extra', type=_str2bool, required=False,
                        help='If True loads separatrix and heating power',
                        default=ddef['extra'])
    parser.add_argument('-sx', '--sharex', type=_str2bool, required=False,
                        help='Should X axis be shared between diag ids?',
                        default=ddef['sharex'], const=True, nargs='?')
    parser.add_argument('-if', '--input_file', type=str, required=False,
                        help='mat file from which to load core_profiles',
                        default=None)
    parser.add_argument('-of', '--output_file', type=str, required=False,
                        help='mat file into which to save synthetic signal',
                        default=None)
    parser.add_argument('-bck', '--background', type=_str2bool, required=False,
                        help='Plot data enveloppe as grey background ?',
                        default=ddef['bck'], const=True, nargs='?')
    return ddef, parser


# #############################################################################
#       Parser dict
# #############################################################################


_DPARSER = {
    'version': parser_version,
    'custom': parser_custom,
    'plot': parser_plot,
    'calc': parser_calc,
}



# -*- coding: utf-8 -*-
"""
Created on Thu May 30 13:38:16 2024

@author: dvezinet
"""

import itertools as itt


import numpy as np
import datastock as ds


from . import _class08_get_data_def
from . import _class08_get_data_vos_broadband
from . import _class08_get_data_vos_spectro


# ##################################################################
# ##################################################################
#                   get data
# ##################################################################


def main(
    coll=None,
    key=None,
    key_cam=None,
    data=None,
    # relevant for LOS data
    segment=None,
    # relevant for spectro data
    rocking_curve=None,
    units=None,
    default=None,
    print_full_doc=None,
    **kwdargs,
):

    # ----------------
    # check inputs
    # ----------------

    (
        key, key_cam,
        spectro, is_vos, is_3d,
        data,
        lok, lcam, lquant,
        davail,
    ) = _check(
        coll=coll,
        key=key,
        key_cam=key_cam,
        data=data,
        rocking_curve=rocking_curve,
        units=units,
        default=default,
        print_full_doc=print_full_doc,
        **kwdargs,
    )

    # print_full_doc
    if key is None:
        return

    # print diag-specific
    if data is None and len(kwdargs) == 0:
        return

    # ----------------
    # extract lists
    # ----------------

    dav = {k0: list(v0['fields'].keys()) for k0, v0 in davail.items()}

    # ----------------
    # build ddata
    # ----------------

    # -----------
    # initialize

    ddata = {}
    dref = {}

    static = True
    daxis = None

    # -------------
    # data quantity

    if data is None or data in lquant:

        # --------------------------
        # data is None => kwdargs

        if data is None:

            # reorder
            ddata = {
                cc: lok[lcam.index(cc)]
                for cc in key_cam if cc in lcam
            }

        # -----------------
        # data in lquant

        elif data in lquant:
            for cc in key_cam:
                dd = coll.dobj['diagnostic'][key]['doptics'][cc][data]
                lc = [
                    isinstance(dd, str) and dd in coll.ddata.keys(),
                ]
                if lc[0]:
                    ddata[cc] = dd
                elif dd is None:
                    pass
                else:
                    msg = f"Unknown data: '{data}'"
                    raise Exception(msg)

        # dref
        dref = {
            k0: coll.ddata[v0]['ref']
            for k0, v0 in ddata.items()
        }

        # units
        if len(ddata) > 0:
            units = coll.ddata[ddata[key_cam[0]]]['units']
        else:
            units = None

        # get actual data
        ddata = {
            k0 : coll.ddata[v0]['data']
            for k0, v0 in ddata.items()
        }

    # ------------------------------------
    # standard LOS-related

    elif data in dav.get('standard - LOS', []):

        if data in ['length', 'tangency_radius', 'alpha']:
            for cc in key_cam:
                ddata[cc], _, dref[cc] = coll.get_rays_quantity(
                    key=key,
                    key_cam=cc,
                    quantity=data,
                    segment=segment,
                    lim_to_segments=False,
                )
            if data in ['length', 'tangency_radius']:
                units = 'm'
            else:
                units = 'rad'

        elif data == 'alpha_pixel':
            for cc in key_cam:

                klos = coll.dobj['diagnostic'][key]['doptics'][cc]['los']
                vectx, vecty, vectz = coll.get_rays_vect(klos)
                dvect = coll.get_camera_unit_vectors(cc)
                sca = (
                    dvect['nin_x'] * vectx[0, ...]
                    + dvect['nin_y'] * vecty[0, ...]
                    + dvect['nin_z'] * vectz[0, ...]
                )

                ddata[cc] = np.arccos(sca)
                dref[cc] = coll.dobj['camera'][cc]['dgeom']['ref']
                units = 'rad'

    # ------------------------------------
    # wavelength-related for spectro diags

    elif data in dav.get('spectro - lamb', []):

        for cc in key_cam:
           ddata[cc], dref[cc] = coll.get_diagnostic_lamb(
               key=key,
               key_cam=cc,
               rocking_curve=rocking_curve,
               lamb=data,
               units=units,
           )
        if data in ['lamb', 'lambmin', 'lambmax', 'dlamb']:
            units = 'm'
        else:
            units = ''

    # ------------------------------
    # data from synthetic diagnostic

    elif data in dav.get('synth', []):

        dref = {}
        daxis = {}
        dsynth = coll.dobj['synth sig'][data]
        for cc in key_cam:
            kdat = dsynth['data'][dsynth['camera'].index(cc)]
            refcam = coll.dobj['camera'][cc]['dgeom']['ref']
            ref = coll.ddata[kdat]['ref']

            c0 = (
                tuple([rr for rr in ref if rr in refcam]) == refcam
                and len(ref) in [len(refcam), len(refcam) + 1]
            )
            if not c0:
                msg = (
                    "Can only plot data that is either:\n"
                    "\t- static: same refs as the camera\n"
                    "\t- has a unique extra dimension\n"
                    "Provided:\n"
                    "\t- refcam: {refcam}\n"
                    "\t- ['{kdat}']['ref']: {ref}"
                )
                raise Exception(msg)

            if len(ref) == len(refcam) + 1:
                static = False
                daxis[cc] = [
                    ii for ii, rr in enumerate(ref) if rr not in refcam
                ][0]

            ddata[cc] = coll.ddata[kdat]['data']
            dref[cc] = ref

            units = coll.ddata[kdat]['units']

    # -----------------
    # raw data

    elif data in dav.get('raw', []):

        ddata = {key_cam[0]: coll.ddata[data]['data']}
        dref = {key_cam[0]: coll.dobj['camera'][key_cam[0]]['dgeom']['ref']}
        units = coll.ddata[data]['units']
        static = True

    # -----------------
    # vos broadband data

    elif data in dav.get('broadband - vos', []):

        ddata, dref, units, static = _class08_get_data_vos_broadband.main(
            coll=coll,
            key=key,
            key_cam=key_cam,
            data=data,
        )

    # -----------------
    # vos spectro data

    elif data in dav.get('spectro - vos', []):

        ddata, dref, units, static = _class08_get_data_vos_spectro.main(
            coll=coll,
            key=key,
            key_cam=key_cam,
            data=data,
        )

    return ddata, dref, units, static, daxis


# ##################################################################
# ##################################################################
#                   check input
# ##################################################################


def _check(
    coll=None,
    key=None,
    key_cam=None,
    data=None,
    rocking_curve=None,
    units=None,
    default=None,
    print_full_doc=None,
    **kwdargs,
):

    # --------------
    # print full doc
    # --------------

    pdef = (
        len(coll.dobj.get('diagnostic')) == 0
        or ((len(coll.dobj.get('diagnostic')) > 1) and key is None)
    )
    print_full_doc = ds._generic_check._check_var(
        print_full_doc, 'print_full_doc',
        types=bool,
        default=pdef,
        allowed=[True] if pdef else [True, False]
    )

    # execute
    if print_full_doc is True:
        davail = _class08_get_data_def.get_davail()
        _print(davail)

    # stop here if relevant
    if pdef is True:
        return [None]*10

    # --------------
    # key, key_cam
    # --------------

    # -------------
    # key, key_cam

    key, key_cam = coll.get_diagnostic_cam(
        key=key,
        key_cam=key_cam,
        default=default,
    )

    # -----------------------
    # spectro, is_vos, is_3d
    # -----------------------

    # -------------
    # spectro

    spectro = coll.dobj['diagnostic'][key]['spectro']

    # -------------
    # is_vos, is_3d

    doptics = coll.dobj['diagnostic'][key]['doptics']
    is_vos = doptics[key_cam[0]].get('dvos') is not None
    is_3d = is_vos and doptics[key_cam[0]]['dvos'].get('indr_3d') is not None

    # ---------------------------
    # get dict of available data
    # ---------------------------

    davail = _class08_get_data_def.get_davail(
        coll=coll,
        key=key,
        key_cam=key_cam,
        # conditions
        spectro=spectro,
        is_vos=is_vos,
        is_3d=is_3d,
    )

    # -----------------------
    # initialize output
    # -----------------------

    lok, lcam, lquant = None, None, None

    # -----------------------
    # preliminary
    # -----------------------

    lc = [data is None, len(kwdargs) == 0]
    if np.sum(lc) < 1:
        msg = (
            "Please provide data xor kwdargs, not both!\n"
            "providing none of them will print the doc\n"
        )
        raise Exception(msg)

    # -----------------------
    # basic checks on data
    # -----------------------

    if all(lc) and print_full_doc is False:

        _print(
            davail,
            key=key,
            spectro=spectro,
            is_vos=is_vos,
            is_3d=is_3d,
        )

    elif data is None:

        # -------------
        # check kwdargs

        dparam = coll.get_param(which='data', returnas=dict)
        lkout = [k0 for k0 in kwdargs.keys() if k0 not in dparam.keys()]

        if len(lkout) > 0:
            msg = (
                "The following args correspond to no data parameter:\n"
                + "\n".join([f"\t- {k0}" for k0 in lkout])
            )
            raise Exception(msg)

        # -----------------------
        # list all available data

        lok = [
            k0 for k0, v0 in coll.ddata.items()
            if v0.get('camera') in key_cam
        ]

        # -------------------
        # Adjust with kwdargs

        if len(kwdargs) > 0:
            lok2 = coll.select(
                which='data', log='all', returnas=str, **kwdargs,
            )
            lok = [k0 for k0 in lok2 if k0 in lok]

        # -----------------------------
        # check there is 1 data per cam

        lcam = [
            coll.ddata[k0]['camera'] for k0 in lok
            if coll.ddata[k0]['camera'] in key_cam
        ]

        if len(set(lcam)) > len(key_cam):
            msg = (
                "There are more / less data identified than cameras:\n"
                f"\t- key_cam:  {key_cam}\n"
                f"\t- data cam: {lcam}\n"
                f"\t- data: {data}"
            )
            raise Exception(msg)

        elif len(set(lcam)) < len(key_cam):
            pass

    else:

        # ----------------
        # allowable values

        lquant = ['etendue', 'amin', 'amax']  # 'los'

        # -------------
        # overall check

        lok = list(itt.chain.from_iterable([
            list(v0['fields'].keys()) for v0 in davail.values()
        ]))

        try:
            data = ds._generic_check._check_var(
                data, 'data',
                types=str,
                allowed=lok,
            )
        except Exception as err:
            msg = (
                f"{err}\n"
                + _print(
                    davail,
                    key=key,
                    spectro=spectro,
                    is_vos=is_vos,
                    is_3d=is_3d,
                    returnas_str=True,
                )
            )
            raise Exception(msg)

    return (
        key, key_cam,
        spectro, is_vos, is_3d,
        data,
        lok, lcam, lquant,
        davail,
    )


# ##################################################################
# ##################################################################
#                   get data doc printing
# ##################################################################


def _print(
    davail,
    key=None,
    spectro=None,
    is_vos=None,
    is_3d=None,
    returnas_str=None,
):

    # -----------------
    # check inputs
    # -----------------

    returnas_str = ds._generic_check._check_var(
        returnas_str, 'returnas_str',
        types=bool,
        default=False,
    )

    # -----------------
    # initialize
    # -----------------

    msg = "\n\n##############################################\n"
    if key is None:
        msg += "The following built-in data is available in general:\n\n"
    else:
        msg += (
            f"The following data is available for '{key}':\n"
            f"- spectro: {spectro}\n"
            f"- is_vos: {is_vos}\n"
            f"- is_3d: {is_3d}\n\n"
        )

    # -----------------
    # loop on davail
    # -----------------

    for k0, v0 in davail.items():

        # header
        msg += f"\n{k0}\n------------------\n"

        # get max length
        ll = [len(k1) for k1 in v0['fields'].keys()]
        nmax = np.max(ll) + 1

        # list of data
        for k1, v1 in v0['fields'].items():
            msg += f"\t- {k1.ljust(nmax)}:  {v1['doc']}\n"

    # -----------------
    # print
    # -----------------

    if returnas_str is True:
        return msg
    else:
        print(msg)
        return
